
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
import logging
import sys
import os.path
from pathlib import Path
import configparser
from holado_logging.common.logging.log_config import LogConfig
from holado_core.common.tools.tools import Tools


logger = logging.getLogger(__name__)

def filter_thread_native_id(record):
    """Inject thread_native_id to log records"""
    from holado_multitask.multitasking.multitask_manager import MultitaskManager
    record.thread_native_id = MultitaskManager.get_thread_id(native=True)
    return record

class LogManager(object):

    def __init__(self):
        self.__loggers_levels = []
        
        self.__config_file_path = None
        
        self.on_console = False
        self.__console_handler = None
        
        # Manage file handlers
        self.in_file = True
        self.__files_lock = threading.Lock()
        self.__file_names = []
        self.__file_handlers = {}
        self.__root_file_name = None
        self.__root_file_handler = None
        self.__root_file_handler_active = False
        
        # Manage file paths LIFO by loggers
        self.__files_by_logger = {}
        
        # Initialize format according python version
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        if MultitaskManager.has_thread_native_id():
            # Exists since python 3.8
            # self.format = '{asctime:s} | {thread_native_id:-5d} | {levelname:5s} | {name:50s} | {message:s}'
            self.format = '{asctime:s} | {process:-5d}-{thread_native_id:-5d} | {levelname:5s} | {name:50s} | {message:s}'
            self.format_short = '{asctime:s} | {message:s}'
            self.style = '{'
        else:
            if sys.version_info > (3, 2):
                # self.format = '{asctime:s} | {thread:-5d} | {levelname:5s} | {module:35s} | {message:s}'
                self.format = '{asctime:s} | {process:-5d}-{thread:-5d} | {levelname:5s} | {name:50s} | {message:s}'
                self.format_short = '{asctime:s} | {message:s}'
                self.style = '{'
            else: 
                self.format = '%(asctime)s | %(process)-5d-%(thread)-5d | %(levelname)5s | %(module)35s | %(message)s'
                self.format_short = '%(asctime)s | %(message)s'
                self.style = '%'
        
    def configure(self):
        self.__config_file_path = LogConfig.config_file_path
        
        if self.__config_file_path:
            config = configparser.ConfigParser()
            config.read(self.__config_file_path)
            
            if config.has_section("loggers_levels"):
                self.__loggers_levels = config.items(section="loggers_levels")
    
    def initialize(self):
        """
        Initialize log manager.
        If log_on_console is True, logs are published on console until a new configuration by calling method set_config
        """
        handlers = []
        if LogConfig.log_on_console:
            self.on_console = True
            self.__console_handler = self.__new_console_handler()
            handlers.append(self.__console_handler)
        
        logging.basicConfig(format=self.format, style=self.style, level=LogConfig.default_level, handlers=handlers)
    
    def has_log_file(self, file_name):
        with self.__files_lock:
            return file_name in self.__file_names or file_name == self.__root_file_name
    
    def set_root_log_file(self, file_name):
        with self.__files_lock:
            if file_name is not None and len(file_name) > 0:
                self.__root_file_name = file_name
                
    def reset_log_files(self):
        with self.__files_lock:
            self.__file_names.clear()
                
    def add_log_file(self, file_name):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if file_name is not None and len(file_name) > 0:
            if file_name == self.__root_file_name:
                raise TechnicalException(f"Log file '{file_name}' is already set as root log file")
            with self.__files_lock:
                if file_name in self.__file_names:
                    raise TechnicalException(f"Log file '{file_name}' is already set")
                self.__file_names.append(file_name)
    
    def remove_log_file(self, file_name):
        with self.__files_lock:
            if file_name is not None and len(file_name) > 0 and file_name in self.__file_names:
                self.__file_names.remove(file_name)
        
    def __new_console_handler(self):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        
        res = logging.StreamHandler()
        res.setFormatter(logging.Formatter(fmt=self.format, style=self.style))
        if MultitaskManager.has_thread_native_id():
            res.addFilter(filter_thread_native_id)
        return res
    
    def set_config(self):
        # print(f"Set logging config: {LogConfig.default_level=} ; {self.on_console=} ; {self.in_file=} ; {self.__root_file_name=}")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Set logging config: {LogConfig.default_level=} ; {self.on_console=} ; {self.in_file=} ; {self.__root_file_name=}")

        # if self.__config_file_path is not None:
        #     logging.config.fileConfig(self.__config_file_path, defaults=None, disable_existing_loggers=False)
        
        logger_ = logging.getLogger()

        # Update log destination to console
        if self.on_console:
            if self.__console_handler is not None:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug("Log destination already set to console.")
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug("Adding log destination to console.")
                self.__console_handler = self.__new_console_handler()
                logger_.addHandler(self.__console_handler)
        else:
            if self.__console_handler is not None:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug("Removing log destination to console.")
                logger_.removeHandler(self.__console_handler)
                self.__console_handler = None
                
        # Update log destination to files
        with self.__files_lock:
            # Remove old log files
            # Note: root file is not removed if it is configured
            for file_name in list(self.__file_handlers.keys()):
                if file_name not in self.__file_names:
                    self.remove_file_handler(file_name, do_remove_log_file=False)
            
            # Add new log files
            if self.__root_file_name and not self.__root_file_handler:
                self.add_root_file_handler()
            for file_name in self.__file_names:
                if file_name not in list(self.__file_handlers.keys()):
                    self.add_file_handler(file_name)
        
        # level
        if logger_.getEffectiveLevel() != LogConfig.default_level:
            logger_.setLevel(LogConfig.default_level)
    
        # Loggers levels
        for name, level in self.__loggers_levels:
            if not name.startswith("#"):
                logging.getLogger(name).setLevel(level)
        
        # WARNING: For local debug only
        # logging.getLogger("holado_logging.common.logging.log_manager").setLevel(logging.DEBUG)
        
    def set_level(self, log_level, do_set_config=True):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if isinstance(log_level, str):
            if hasattr(logging, log_level):
                log_level = getattr(logging, log_level)
            else:
                raise TechnicalException(f"Unexpected log level string '{log_level}'")
            
        LogConfig.default_level = log_level
        if do_set_config:
            self.set_config()
        
    def add_root_file_handler(self):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        logger_ = logging.getLogger()
            
        if not self.__root_file_name:
            raise TechnicalException("Root log file is not defined")
            
        if self.__root_file_handler is None:
            logger.info("Creating file handler to root file '{}'.".format(self.__root_file_name))
            self.__root_file_handler = self.__new_file_handler(self.__root_file_name)
            
        logger_.addHandler(self.__root_file_handler)
        self.__root_file_handler_active = True
        
    def remove_root_file_handler(self, do_reset=False):
        logger_ = logging.getLogger()
        if self.__root_file_handler:
            logger.info(f"Removing log destination to root file '{self.__root_file_name}'.")
            logger_.removeHandler(self.__root_file_handler)
            self.__root_file_handler_active = False
            if do_reset:
                self.__root_file_handler = None
        
    def add_file_handler(self, file_name, logger_=None):
        if logger_ is None:
            logger_ = logging.getLogger()
            
        # In case this method is called outside "set_config" method, add file_name as a configured log_file
        if not self.has_log_file(file_name):
            self.add_log_file(file_name)
            
        if file_name in self.__file_handlers:
            logger.debug("Log destination already set to file '{}'.".format(file_name))
        else:
            logger.info("Adding log destination to file '{}'.".format(file_name))
            file_handler = self.__new_file_handler(file_name)
            self.__file_handlers[file_name] = file_handler
            self.add_existing_file_handler_to_logger(file_name, logger_)
    
    def __new_file_handler(self, file_path, use_format_short=False):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        
        Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)
        
        res = logging.FileHandler(file_path, mode='w', encoding='utf8')
        fmt = self.format_short if use_format_short else self.format
        res.setFormatter(logging.Formatter(fmt=fmt, style=self.style))
        if MultitaskManager.has_thread_native_id():
            res.addFilter(filter_thread_native_id)
        
        return res
    
    def remove_file_handler(self, file_name, logger_=None, do_remove_log_file=True):
        if logger_ is None:
            logger_ = logging.getLogger()
            
        logger.info("Removing log destination to file '{}'.".format(file_name))
        self.remove_existing_file_handler_from_logger(file_name, logger_)
        del self.__file_handlers[file_name]
        
        if do_remove_log_file:
            self.remove_log_file(file_name)
    
    def add_existing_file_handler_to_logger(self, file_name, logger_):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if self.__root_file_name == file_name:
            file_handler = self.__root_file_handler
        else:
            if file_name not in self.__file_handlers:
                raise TechnicalException(f"Not set log destination to file '{file_name}'")
            file_handler = self.__file_handlers[file_name]
        logger_.addHandler(file_handler)
    
    def remove_existing_file_handler_from_logger(self, file_name, logger_):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if self.__root_file_name == file_name:
            file_handler = self.__root_file_handler
        else:
            if file_name not in self.__file_handlers:
                raise TechnicalException(f"Not set log destination to file '{file_name}'")
            file_handler = self.__file_handlers[file_name]
        logger_.removeHandler(file_handler)
    
    def __remove_existing_file_handlers(self, log_level_if_exists=None):
        with self.__files_lock:
            if len(self.__file_names) > 0:
                if log_level_if_exists is not None:
                    logger_ = logging.getLogger()
                    logger_.log(log_level_if_exists, f"Removing existing file handlers: {self.__file_names}")
                file_names = list(self.__file_names)
            else:
                file_names = []
        
        # Note: removes are done outside lock to avoid a deadlock
        for fn in file_names:
            self.remove_file_handler(fn)
    
    def enter_log_file(self, file_name, do_remove_root_file_handler=True, do_remove_other_file_handlers=True):
        if do_remove_other_file_handlers:
            self.__remove_existing_file_handlers(log_level_if_exists=logging.ERROR)
        
        self.add_file_handler(file_name)
        
        if do_remove_root_file_handler:
            self.remove_root_file_handler(do_reset=False)
            
    def leave_log_file(self, file_name, do_remove_log_file=True):
        if not self.__root_file_handler_active:
            self.add_root_file_handler()
        self.remove_file_handler(file_name, do_remove_log_file=do_remove_log_file)
    
    def enter_log_file_for_logger(self, logger_, file_path, use_format_short=False, switch_in=True):
        logger_name = logger_.name
        if logger_name not in self.__files_by_logger:
            self.__files_by_logger[logger_name] = []
        
        if switch_in and len(self.__files_by_logger[logger_name]) > 0:
            logger_.removeHandler(self.__files_by_logger[logger_name][-1][1])
        
        file_handler = self.__new_file_handler(file_path, use_format_short=use_format_short)
        logger_.addHandler(file_handler)
        self.__files_by_logger[logger_name].append( (file_path, file_handler) )
        
    def leave_log_file_for_logger(self, logger_, switch_out=True):
        logger_name = logger_.name
        _, file_handler = self.__files_by_logger[logger_name].pop()
        logger_.removeHandler(file_handler)
        
        if switch_out and len(self.__files_by_logger[logger_name]) > 0:
            logger_.addHandler(self.__files_by_logger[logger_name][-1][1])
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_python.common.tools.datetime import DateTime
import re
from datetime import datetime, timedelta
from holado_protobuf.ipc.protobuf.abstracts.type import Type
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


#TODO: add all types of https://googleapis.dev/python/protobuf/latest/google/protobuf.html
 

class Duration(Type):
    """
    Manage actions with type google.protobuf.Duration
    """
    @classmethod
    def protobuf_class(cls):
        from google.protobuf.duration_pb2 import Duration as gp_Duration  # @UnresolvedImport
        return gp_Duration
    
    @classmethod
    def _set_object_value(cls, obj, value, raise_exception=False):
        if isinstance(value, cls.protobuf_class()):
            obj.CopyFrom(value)
        elif isinstance(value, timedelta):
            obj.FromTimedelta(value)
        elif isinstance(value, int) or isinstance(value, float):
            nanos = round(value * 1e9)
            obj.FromNanoseconds(nanos)
        elif isinstance(value, str):
            if re.match(r"-?\d+(?:.\d+)?s", value): 
                obj.FromJsonString(value)
            else:
                if raise_exception:
                    raise FunctionalException(f"For Duration objects, the value must be in a string format '-?\\d+(?:.\\d+)?s' (ex: '1s', '1.01s', '1.0000001s', '-3.100s'). Got value: [{value}] (type: {Typing.get_object_class_fullname(value)})")
                else:
                    return False
        else:
            if raise_exception:
                raise FunctionalException(f"For Duration objects, the value must be a timedelta, an int or a string. Got value: [{value}] (type: {Typing.get_object_class_fullname(value)})")
            else:
                return False
        return True


class Message(Type):
    """
    Manage actions with type google.protobuf.Message
    """
    @classmethod
    def protobuf_class(cls):
        from google.protobuf.message import Message as gp_Message
        return gp_Message
    
    @classmethod
    def _set_object_value(cls, obj, value, raise_exception=False):
        if value is None:
            obj.SetInParent()
        elif isinstance(value, cls.protobuf_class()):
            obj.CopyFrom(value)
        elif isinstance(value, dict):
            from google.protobuf import json_format
            json_format.ParseDict(value, obj)
        else:
            if raise_exception:
                raise FunctionalException(f"For Message objects, the value must be None or a {cls.protobuf_class()}. Got value of type {Typing.get_object_class_fullname(value)}")
            else:
                return False
        return True
        

class Timestamp(Type):
    """
    Manage actions with type google.protobuf.Timestamp
    """
    @classmethod
    def protobuf_class(cls):
        from google.protobuf.timestamp_pb2 import Timestamp as gp_Timestamp  # @UnresolvedImport
        return gp_Timestamp
    
    @classmethod
    def _set_object_value(cls, obj, value, raise_exception=False):
        if isinstance(value, cls.protobuf_class()):
            obj.CopyFrom(value)
        elif isinstance(value, datetime):
            obj.FromDatetime(value)
        elif isinstance(value, str):
            if re.match(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:.\d+)?Z", value): 
                obj.FromJsonString(value)
            elif DateTime.is_str_datetime(value):
                dt = DateTime.str_2_datetime(value)
                obj.FromDatetime(dt)
            else:
                if raise_exception:
                    raise FunctionalException(f"For Timestamp objects, the value must be in a string format like '2022-01-01T01:00:00Z' or '2022-01-01 01:00:00'. Got value: [{value}] (type: {Typing.get_object_class_fullname(value)})")
                else:
                    return False
        else:
            if raise_exception:
                raise FunctionalException(f"For Timestamp objects, the value must be a datetime or a string. Got value: [{value}] (type: {Typing.get_object_class_fullname(value)})")
            else:
                return False
        return True
        

    

    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_test.scenario.scenario_tools import ScenarioStatusInfo,\
    ScenarioTools



logger = logging.getLogger(__name__)


class BehaveScenarioTools(ScenarioTools):
    """
    Gives usefull tools for scenarios.
    """
    

    @classmethod
    def get_scenario_uid(cls, scenario):
        return f"{scenario.filename} at l.{scenario.line}"
    
    @classmethod
    def format_scenario_short_description(cls, scenario):
        return f"{scenario.filename} at l.{scenario.line}"
    
    @classmethod
    def compute_validation_status(cls, scenario, scenario_context, step_failed):
        if step_failed is not None and hasattr(scenario_context, "is_in_preconditions") and scenario_context.is_in_preconditions:
            return "Failed in Preconditions"
        elif step_failed is not None and step_failed.keyword == "Given":
            return "Failed in Given"
        elif step_failed is not None or scenario.status.has_failed():
            return "Failed"
        elif hasattr(scenario, 'sut_failed') and scenario.sut_failed:
            return "Failed in SUT"
        else:
            return scenario.status.name.capitalize()
        
    @classmethod
    def get_step_failed_info(cls, scenario):
        res_step, res_step_number = None, None
        for ind, step in enumerate(scenario.steps):
            if step.status.has_failed():
                res_step, res_step_number = step, ind+1
                break
        return res_step, res_step_number
    





# -*- coding: utf-8 -*-


from test_holado.steps.public_steps import *

# Note: following lines are commented since, with this example files architecture, these steps are automatically found and imported by behave
# from steps.config_steps import *







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from typing import NamedTuple
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.block.scope_function import ScopeFunction
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class BlockManager(object):
    '''
    Manage blocks.
    
    Currently it manages definition of scopes (functions and loops).
    '''
    
    def __init__(self):
        self.__scope_by_id = {}
        self.__scope_define_by_thread = {}
        
    @property
    def is_in_define(self):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        thread_id = MultitaskManager.get_thread_id()
        return thread_id in self.__scope_define_by_thread
    
    @property
    def is_in_sub_define(self):
        return self.__get_scope_define().level > 1
    
    @property
    def scope_in_define(self):
        if not self.is_in_define:
            raise FunctionalException("No block is under define")
        scope_id = self.__get_scope_define().scope_id
        return self.__scope_by_id[scope_id]
    
    def begin_define_scope(self, scope_type, scope_name, block_scope_type=None):
        """
        Begin define of a scope of given scope_name.
        Parameter block_scope_type specifies the scope class to use (any class inheriting from BaseScope ; default: ScopeFunction).
        """
        if block_scope_type is None:
            block_scope_type = ScopeFunction
            
        if self.is_in_define:
            raise FunctionalException("A block is already under define")

        scope_id = self.__get_scope_ID(scope_name)
        if scope_id in self.__scope_by_id:
            raise FunctionalException(f"Scope '{scope_name}' (ID: {scope_id}) was already defined")
        
        block_scope = block_scope_type(ID=scope_id, name=scope_name)
        
        self.__new_scope_define(scope_type, scope_id, scope_name)
        self.__scope_by_id[scope_id] = block_scope
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[BlockManager] Define of '{scope_name}' (ID: {scope_id}) => begin")
        
    def end_define_scope(self, scope_type):
        if not self.is_in_define:
            raise TechnicalException("No block is under define")
        
        scope_define = self.__get_scope_define()
        if scope_define.scope_type != scope_type:
            raise TechnicalException(f"Scope under define is '{scope_define.scope_type}' (expected: '{scope_type}')")
        
        self.__delete_scope_define(scope_define.scope_id)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[BlockManager] Define of '{scope_define.scope_name}' (ID: {scope_define.scope_id}) => end")

    def increase_define_level(self):
        self.__get_scope_define().level += 1

    def decrease_define_level(self):
        self.__get_scope_define().level -= 1
        
    def has_scope(self, scope_name):
        scope_id = self.__get_scope_ID(scope_name)
        return scope_id in self.__scope_by_id

    def delete_scope(self, scope_name=None, scope_id=None):
        if scope_id is None:
            scope_id = self.__get_scope_ID(scope_name)
        
        # Check scope
        self.__check_block_is_not_under_define(scope_id)
        self.__check_block_is_defined(scope_id)
        
        # Delete scope
        del self.__scope_by_id[scope_id]

    def __check_block_is_not_under_define(self, block_id):
        scope_define = self.__get_scope_define(raise_exception=False)
        if scope_define is not None and scope_define.scope_id == block_id:
            raise FunctionalException(f"Block '{self.scope_in_define.name}' is under define")

    def __check_block_is_defined(self, block_id):
        if block_id not in self.__scope_by_id:
            raise FunctionalException(f"Block of ID '{block_id}' is not define")

    def __get_scope_ID(self, scope_name):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        return f"SCOPE[{MultitaskManager.get_thread_id()};{scope_name}]"
    
    def __get_scope_define(self, thread_id=None, raise_exception=True):
        from holado_multitask.multitasking.multitask_manager import MultitaskManager
        if thread_id is None:
            thread_id = MultitaskManager.get_thread_id()
        if thread_id not in self.__scope_define_by_thread:
            if raise_exception:
                raise TechnicalException(f"No scope is under define for thread {thread_id}")
            else:
                return None
        return self.__scope_define_by_thread[thread_id]

    def __new_scope_define(self, scope_type, scope_id, scope_name, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()

        res = NamedTuple("ScopeDefine", scope_type=str, scope_id=str, scope_name=str, level=int)
        res.scope_type = scope_type
        res.scope_id = scope_id
        res.scope_name = scope_name
        res.level = 1
        
        self.__scope_define_by_thread[thread_id] = res
        return res

    def __delete_scope_define(self, scope_id, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        if self.__get_scope_define(thread_id).scope_id != scope_id:
            raise TechnicalException(f"For thread {thread_id}, scope under define has ID '{self.__get_scope_define(thread_id).scope_id}' (expected: {scope_id})")
        del self.__scope_define_by_thread[thread_id]
    
    def process_scope(self, scope_name=None, scope_id=None, *args, **kwargs):
        res = None
        if scope_id is None:
            scope_id = self.__get_scope_ID(scope_name)
        
        # Check scope
        self.__check_block_is_not_under_define(scope_id)
        self.__check_block_is_defined(scope_id)
        
        # Start scope
        func = self.__scope_by_id[scope_id]
        try:
            res = func.process(*args, **kwargs)
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            raise TechnicalException(exc)
        
        return res




from builtins import super
from holado.common.context.session_context import SessionContext
import logging

logger = logging.getLogger(__name__)


class TestServerSessionContext(SessionContext):
    
    def __init__(self):
        super().__init__("TestServerSession")
        
    def configure(self, session_kwargs=None):
        logger.info("Configuring TestServerSessionContext")
        
        super().configure(session_kwargs)
        
        # Override default registered modules
         

        
        # Register new modules
        
        from holado_test.tools.test_server.server.core.server_manager import TestServerManager
        self.services.register_service_type("server_manager", TestServerManager,
                            lambda m: m.initialize(self.resource_manager) )
        
        
        
    def initialize(self, session_kwargs=None):
        if session_kwargs is None:
            session_kwargs = {}
        # raise_if_not_exist = session_kwargs.get("raise_if_not_exist", True)
        # do_import = session_kwargs.get("import_compiled_proto", True)
        
        # Call default initialization
        super().initialize(session_kwargs)
        
        
        
        
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class ScopeManager(object):
    '''
    Manage scopes.
    '''
    
    def __init__(self):
        self.__scope_info_by_thread_and_type = {}
    
    
    # Manage scope info
    
    def scope_info(self, scope_type, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        self.__ensure_scope_info_exists(scope_type, thread_id)
        return self.__scope_info_by_thread_and_type[thread_id][scope_type]
    
    def __check_scope_info_exists(self, scope_type, thread_id, raise_exception=True):
        res = thread_id in self.__scope_info_by_thread_and_type and scope_type in self.__scope_info_by_thread_and_type[thread_id]
        if not res and raise_exception:
            raise FunctionalException(f"Scope info doesn't exist (thread: {thread_id} ; scope type: {scope_type})")
        return res
    
    def __ensure_scope_info_exists(self, scope_type, thread_id):
        if thread_id not in self.__scope_info_by_thread_and_type:
            self.__scope_info_by_thread_and_type[thread_id] = {}
        if scope_type not in self.__scope_info_by_thread_and_type[thread_id]:
            self.__scope_info_by_thread_and_type[thread_id][scope_type] = {}
    
    
    # Manage scope levels
    
    def has_scope_level(self, scope_type, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        return self.__check_scope_level_exists(scope_type, thread_id, raise_exception=False)
    
    def scope_level(self, scope_type, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        self.__ensure_scope_level_exists(scope_type, thread_id)
        return self.scope_info(scope_type, thread_id)["level"]
    
    def __check_scope_level_exists(self, scope_type, thread_id, raise_exception=True):
        res = self.__check_scope_info_exists(scope_type, thread_id, raise_exception) and "level" in self.scope_info(scope_type, thread_id)
        if not res and raise_exception:
            raise FunctionalException(f"Scope level doesn't exist (thread: {thread_id} ; scope type: {scope_type})")
        return res
    
    def __ensure_scope_level_exists(self, scope_type, thread_id):
        self.__ensure_scope_info_exists(scope_type, thread_id)
        if "level" not in self.__scope_info_by_thread_and_type[thread_id][scope_type]:
            self.__scope_info_by_thread_and_type[thread_id][scope_type]["level"] = 0
        
    def reset_scope_level(self, scope_type=None, level=0, thread_id=None):
        """
        Reset a scope level to given level.
        If scope_type is None, all scope types of the thread are reset.
        """
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
            
        if scope_type is None:
            if thread_id in self.__scope_info_by_thread_and_type:
                for s_type in self.__scope_info_by_thread_and_type[thread_id]:
                    self.reset_scope_level(s_type, level, thread_id)
        else:
            if self.scope_level(scope_type, thread_id) != level:
                self.__scope_info_by_thread_and_type[thread_id][scope_type]["level"] = level
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Scope level reset to {level} (thread: {thread_id} ; scope type: {scope_type})")

    def increase_scope_level(self, scope_type, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        self.__ensure_scope_level_exists(scope_type, thread_id)
        
        self.__scope_info_by_thread_and_type[thread_id][scope_type]["level"] += 1
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Scope level increased to {self.__scope_info_by_thread_and_type[thread_id][scope_type]['level']} (thread {thread_id} ; scope type: {scope_type})")

    def decrease_scope_level(self, scope_type, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        self.__ensure_scope_level_exists(scope_type, thread_id)
        
        self.__scope_info_by_thread_and_type[thread_id][scope_type]["level"] -= 1
        if self.__scope_info_by_thread_and_type[thread_id][scope_type]["level"] < 0:
            raise TechnicalException(f"Scope level decreased under 0 (thread {thread_id} ; scope type: {scope_type})")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Scope level decreased to {self.__scope_info_by_thread_and_type[thread_id][scope_type]['level']} (thread {thread_id} ; scope type: {scope_type})")
    
    
    # Manage conditions scopes
    
    def log_conditions_info(self, prefix_message, thread_id=None):
        scope_info = self.scope_info("conditions", thread_id)
        if not prefix_message.endswith("."):
            prefix_message += "."
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"{prefix_message} Conditions info: {scope_info}")
    
    def increase_condition_level(self, thread_id=None):
        self.increase_scope_level("conditions", thread_id)
        
        scope_info = self.scope_info("conditions", thread_id)
        if "levels_data" not in scope_info:
            scope_info["levels_data"] = []
        if len(scope_info["levels_data"]) != scope_info["level"] - 1:
            raise TechnicalException(f"levels_data has size {len(scope_info['levels_data'])} whereas level is {scope_info['level']}")
        scope_info["levels_data"].append({"is_valid":False, "had_valid":False})
        
        self.log_conditions_info("Increased condition level")
    
    def decrease_condition_level(self, thread_id=None):
        self.decrease_scope_level("conditions", thread_id)
        
        scope_info = self.scope_info("conditions", thread_id)
        if len(scope_info["levels_data"]) != scope_info["level"] + 1:
            raise TechnicalException(f"levels_data has size {len(scope_info['levels_data'])} whereas level is {scope_info['level']}")
        scope_info["levels_data"].pop()
        
        self.log_conditions_info("Decreased condition level")
        
    def is_in_condition(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        return self.__check_scope_level_exists("conditions", thread_id, raise_exception=False) \
            and self.scope_level("conditions", thread_id) > 0
        
    def is_in_valid_condition(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        res = False
        if self.is_in_condition(thread_id):
            scope_info = self.scope_info("conditions", thread_id)
            res = scope_info["levels_data"][-1]["is_valid"]
        return res
        
    def had_valid_condition(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        res = False
        if self.is_in_condition(thread_id):
            scope_info = self.scope_info("conditions", thread_id)
            res = scope_info["levels_data"][-1]["had_valid"]
        return res
        
    def enter_in_valid_condition(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        if not self.is_in_condition(thread_id):
            raise TechnicalException(f"Not in a condition (conditions info: {self.scope_info('conditions', thread_id)})")
        
        scope_info = self.scope_info("conditions", thread_id)
        scope_info["levels_data"][-1]["is_valid"] = True
        
    def leave_condition(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        if not self.is_in_condition(thread_id):
            raise TechnicalException(f"Not in a condition (conditions info: {self.scope_info('conditions', thread_id)})")
        
        scope_info = self.scope_info("conditions", thread_id)
        if scope_info["levels_data"][-1]["is_valid"]:
            scope_info["levels_data"][-1]["is_valid"] = False
            scope_info["levels_data"][-1]["had_valid"] = True


    # Manage steps scope
    
    def set_step_context(self, step_context, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        
        scope_info = self.scope_info("steps")
        if "levels_data" not in scope_info:
            scope_info["levels_data"] = []

        step_level = self.scope_level("steps")
        if step_level + 1 < len(scope_info["levels_data"]):
            scope_info["levels_data"] = scope_info["levels_data"][0:step_level+1]
        for level in range(step_level+1):
            if level + 1 > len(scope_info["levels_data"]):
                scope_info["levels_data"].append({})
                
        scope_info["levels_data"][step_level]["step_context"] = step_context

    def get_step_context(self, thread_id=None):
        if thread_id is None:
            from holado_multitask.multitasking.multitask_manager import MultitaskManager
            thread_id = MultitaskManager.get_thread_id()
        
        scope_info = self.scope_info("steps")
        if "levels_data" not in scope_info:
            raise TechnicalException(f"No step level is registered")

        step_level = self.scope_level("steps")
        if step_level >= len(scope_info["levels_data"]):
            raise TechnicalException(f"Step level {step_level} is not registered")
        if "step_context" not in scope_info["levels_data"][step_level]:
            raise TechnicalException(f"No step context is registered for level {step_level}")
            
        return scope_info["levels_data"][step_level]["step_context"]




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
import csv
from holado_helper.script.behave_action_script import BehaveActionScript, BehaveActionScriptFile
from holado_helper.script.job import Job
from holado.common.context.session_context import SessionContext

logger = logging.getLogger(__name__)


class CSVActionScriptFile(BehaveActionScriptFile):
    def __init__(self):
        super().__init__(file_ext='.csv')
        self.__text_interpreter = SessionContext.instance().text_interpreter

    def read_jobs(self, file_path, action_script):
        action = action_script.args.action
        
        list_action_params = self._read_action_params_from_csv_file(file_path)
        
        actions = []
        for index, ap in enumerate(list_action_params):
            behave_tags = action_script.action_registry.get_action_behave_tags(action)
            behave_args = action_script._build_behave_args(tags=behave_tags)
            static_params = action_script.action_registry.get_action_static_parameters_values(action)
            action_info = action_script.new_action_info(action, behave_args=behave_args, params=ap, static_params=static_params, index=index)
            actions.append(action_info)
        
        job = Job(actions=actions, in_parallel=False)
        
        return [job]
    
    def _read_action_params_from_csv_file(self, file_path):
        res = []
        
        with open(file_path, 'r') as fin:
            dr = csv.DictReader(fin, delimiter=';', quoting=csv.QUOTE_NONE)
            
            for row in dr:
                params_csv = dict(row)
                
                # Interpret parameter values to get expected values instead of strings
                params = {}
                for key, value in params_csv.items():
                    val_to_interpret = f"${{{value}}}"
                    val_interpreted = self.__text_interpreter.interpret(val_to_interpret)
                    if val_interpreted != val_to_interpret:
                        params[key] = val_interpreted
                    else:
                        params[key] = value
                
                res.append(params)
        
        return res

    def write_jobs(self, file_path, jobs, action_script):
        list_action_infos = []
        for job in jobs:
            list_action_infos.extend(job.actions)
        
        self._write_action_params_in_csv_file(file_path, list_action_infos)
    
    def _write_action_params_in_csv_file(self, file_path, list_action_infos):
        # Define fieldnames
        if len(list_action_infos) < 1:
            return
        fieldnames = list(list_action_infos[0].params.keys())
        
        with open(file_path, 'w') as fout:
            dw = csv.DictWriter(fout, fieldnames=fieldnames, delimiter=';', quoting=csv.QUOTE_NONE)
            dw.writeheader()
            for action_info in list_action_infos:
                dw.writerow(action_info.params)



class CSVActionScript(BehaveActionScript):
    def __init__(self, work_dir_path=None, actor=None, action_registry=None, script_file=None):
        if script_file is None:
            script_file = CSVActionScriptFile()
        super().__init__(work_dir_path=work_dir_path, actor=actor, action_registry=action_registry, script_file=script_file)
    
    def get_description(self, script_command, short_description, action):
        return """{short_description}
        
        Usage:
            1. Generate a template file for the desired action with parameter '--output-template' as in example 1.
                It is possible to specify the name of the output file with parameter '-o'.
            2. Copy the template line in order to have one line per action to do, and modify the template file with desired values. 
            3. Process actions described in a CSV file with parameter '-i'.
                If an error occurs at a line, execution is stopped.
            4. See result in output file. If parameter '-o' is not specified, the output filename is generated from parameter '-i'.
        
        In template files, a default value is set in each column. 
        For columns with enum values, the default value is a combination of each possible enum values separated by a '|', but only one value is possible.
        
        
        In input and output files, the first column 'STATUS' describes the status of execution of each line:
            - 'TODO' (or ''): the line is to process (in output file, it means that execution was stopped on an error with a previous line)
            - 'SUCCESS': the line was in success
            - 'ERROR': the line was in error
            - others: the line is not processed, you can specify your own status to skip this line
        In template files, this column is automatically initialized with status 'TODO'.
        Any output file can be used as input file, the lines that are in status 'ERROR' or 'TODO' will be processed again.
        
        Example 1: Output CSV template for action '{action}'
                
                {script_command} {action} --output-template -o {action}.csv
                
        Example 2: Process action '{action}' with given input CSV
                
                {script_command} {action} -i {action}.csv
        """.format(short_description=short_description, script_command=script_command, action=action)
    
    def _new_argument_parser(self, description=None, additional_arguments=None, add_help=True):
        parser = super()._new_argument_parser(description=description, additional_arguments=additional_arguments, add_help=add_help)
        
        # Add mode serie/parallel
        parser.add_argument('-m', '--mode', choices=['serial', 'parallel'], dest='mode', default='serial', 
                     help='Define if CSV rows must be processed in serial mode or in parallel mode')
        
        return parser
    






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tools.tools import Tools
from holado_yaml.yaml.yaml_client import YAMLClient
from holado_core.common.exceptions.technical_exception import TechnicalException


logger = logging.getLogger(__name__)

try:
    import yaml  # @UnresolvedImport
    with_yaml = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"YAML is not available with pyyaml. Initialization failed on error: {exc}")
    with_yaml = False



class PyYAMLClient(YAMLClient):
    """
    Client for actions on YAML files.
    """
    
    @classmethod
    def is_available(cls):
        return with_yaml
    
    def __init__(self, name=None, client_type=None):
        super().__init__(name)
        
        if client_type == "base":
            self.__loader_type = yaml.BaseLoader
            self.__dumper_type = yaml.BaseDumper
        elif client_type == "full":
            self.__loader_type = yaml.FullLoader
            self.__dumper_type = yaml.Dumper
        elif client_type == "safe" or client_type is None:
            self.__loader_type = yaml.SafeLoader
            self.__dumper_type = yaml.SafeDumper
        elif client_type == "unsafe":
            self.__loader_type = yaml.UnsafeLoader
            self.__dumper_type = yaml.Dumper
        else:
            raise TechnicalException(f"Unmanaged client type '{client_type}'")
    
    def load_io_file(self, file_like_object):
        return yaml.load(file_like_object, self.__loader_type)
    
    def load_multiple_documents_io_file(self, file_like_object):
        return list(yaml.load_all(file_like_object, self.__loader_type))
        
    def save_in_io_file(self, file_like_object, data, **kwargs):
        if 'sort_keys' not in kwargs:
            kwargs['sort_keys'] = False
        yaml.dump(data, file_like_object, **kwargs)
    







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_report.report.builders.report_builder import ReportBuilder
from holado_system.system.filesystem.file import File
from holado_python.common.tools.datetime import DateTime, FORMAT_DATETIME_ISO
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_test.common.context.context_tools import ContextTools

logger = logging.getLogger(__name__)



class SummaryScenarioByCategoryReportBuilder(ReportBuilder):
    def __init__(self, filepath, file_format='txt', exclude_statuses=None, exclude_categories=None, with_scenario_end_date=True, with_scenario_period=False, with_step_period=False, use_compact_format=True):
        self.__file_format = file_format.lower()
        self.__exclude_statuses = exclude_statuses
        self.__exclude_categories = exclude_categories
        self.__with_scenario_end_date = with_scenario_end_date
        self.__with_scenario_period = with_scenario_period
        self.__with_step_period = with_step_period
        self.__use_compact_format = use_compact_format
        
        if self.__file_format == 'txt':
            self.__file = File(filepath, auto_flush=False, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt')")
        
        self.__scenarios_by_category = {}
        self.__categories_order = [
                'Newly Failed',
                'Regression',
                'Always Failed',
                'Random',
                'Newly Failed but Not Relevant',
                'Regression but Not Relevant',
                'Always Not Relevant',
                'Random but Not Relevant',
                'Fixed',
                'Always Success',
                'Unknown'
            ]
        
    def build(self):
        '''
        The file is built after each scenario
        '''
        pass
        
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        status_info = ReportManager.ScenarioTools.get_current_scenario_status_info(scenario)
        if status_info.validation_category is not None:
            ind = status_info.validation_category.find(' (')
            category = status_info.validation_category[:ind] if ind > 0 else status_info.validation_category
        else:
            category = None
        
        # Manage excluded scenario
        if self.__exclude_statuses and status_info.validation_status in self.__exclude_statuses:
            return
        if self.__exclude_categories and category in self.__exclude_categories:
            return
        
        if category is not None:
            # Add scenario information into category
            category_str = f" => {status_info.validation_category}" if status_info.validation_category else ""
            scenario_prefix_str = f"{ContextTools.format_context_period(status_info.scenario_context)} " if self.__with_scenario_period \
                                  else f"{DateTime.datetime_2_str(status_info.scenario_context.end_datetime, FORMAT_DATETIME_ISO)} - " if self.__with_scenario_end_date \
                                  else ""
            if status_info.step_failed:
                step_format_kwargs = {'step_context': status_info.step_context if self.__with_step_period else None,
                                      'dt_ref': status_info.scenario_context.start_datetime if self.__use_compact_format else None}
                scenario_txt = f"{scenario_prefix_str}{ReportManager.ScenarioTools.format_scenario_short_description(scenario)} - {ReportManager.StepTools.format_step_short_description(status_info.step_failed, status_info.step_failed_nb, has_failed=True, **step_format_kwargs)} - {status_info.validation_status}{category_str}"
            else:
                scenario_txt = f"{scenario_prefix_str}{ReportManager.ScenarioTools.format_scenario_short_description(scenario)} - {status_info.validation_status}{category_str}"
            
            if category not in self.__scenarios_by_category:
                self.__scenarios_by_category[category] = []
            self.__scenarios_by_category[category].append(scenario_txt)
            
            # Update categories order with unexpected category
            if category not in self.__categories_order:
                self.__categories_order.append(category)
            
            self.__update_file()
        
    def __update_file(self):
        with self.__file as fout:
            for category in self.__categories_order:
                if category in self.__scenarios_by_category:
                    fout.writelines([
                        f"## {category}",
                        ""
                        ])
                    fout.writelines(self.__scenarios_by_category[category])
                    fout.writelines([
                        ""
                        ""
                        ])
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import os
from holado_core.common.tools.tools import Tools
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)


try:
    from mss import mss  # @UnresolvedImport
    with_mss = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"MssManager is not available. Initialization failed on error: {exc}")
    with_mss = False

class MssManager(object):
    """ Class managing MSS library.
    
    It is implemented internally with 'mss' package (https://python-mss.readthedocs.io/usage.html)
    """
    
    @classmethod
    def is_available(cls):
        return with_mss
    
    @classmethod
    def make_monitors_screenshots(cls, destination_path, context_description):
        """
        Make screenshots for debug (monitors, current UI in focus).
        @param destination_path Destination path
        @param context_description Context description that will be inserted in file names
        """
        screenshotter = mss()
        
        # Create file prefix
        date_str = DateTime.now().strftime('%Y-%m-%d_%H-%M-%S.%f')
        file_prefix = f"{date_str}-{context_description}-monitor"
        screenshot_path = os.path.join(destination_path, file_prefix)
        
        # Take screenshots
        for nb in range(1, len(screenshotter.monitors)):
            screen_path = f"{screenshot_path}{nb}.png"
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Making monitor screenshot of monitor {nb} in '{screen_path}'")
            screenshotter.save(screen_path, nb)
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import super
from holado_core.common.tables.comparators.table_cell_comparator import TableCellComparator
from holado_value.common.tools.value_types import ValueTypes
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.verify_exception import VerifyException
from holado.common.context.session_context import SessionContext
from holado_core.common.tools.comparators.comparator import CompareOperator
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.tools.comparators.object_comparator import ObjectComparator
from holado_value.common.tables.value_table_cell import ValueTableCell
from holado_scripting.text.verifier.text_verifier import TextVerifier
from holado_core.common.tools.tools import Tools
from holado_python.common.tools.comparators.type_comparator import TypeComparator
from holado_python.common.tools.comparators.string_comparator import StringComparator
from holado_python.standard_library.typing import Typing
from holado_python.common.tools.comparators.float_comparator import FloatComparator
from holado.common.handlers.undefined import undefined_value

logger = logging.getLogger(__name__)


class Table2ValueTable_CellComparator(TableCellComparator):
    def __init__(self, **kwargs):
        super().__init__()
        
        self.__kwargs = kwargs
        self.__float_comparator = None
    
    def __get_float_comparator(self):
        if self.__float_comparator is None:
            float_kwargs = {}
            if 'float_diff_precision' in self.__kwargs:
                float_kwargs['diff_precision'] = self.__kwargs['float_diff_precision']
            if 'float_relative_precision' in self.__kwargs:
                float_kwargs['relative_precision'] = self.__kwargs['float_relative_precision']
            self.__float_comparator = FloatComparator(**float_kwargs)
        return self.__float_comparator
    
    def __get_scenario_context(self):
        return SessionContext.instance().get_scenario_context()
            
    def __get_text_verifier(self) -> TextVerifier:
        return self.__get_scenario_context().get_text_verifier()
    
    def __get_variable_manager(self):
        return self.__get_scenario_context().get_variable_manager()
    
    @property
    def __type_comparator(self):
        if not hasattr(self, "__type_comparator_inst"):
            self.__type_comparator_inst = TypeComparator()
        return self.__type_comparator_inst

    @property
    def __string_comparator(self):
        if not hasattr(self, "__string_comparator_inst"):
            self.__string_comparator_inst = StringComparator()
            self.__string_comparator_inst.do_convert_input1 = False
        return self.__string_comparator_inst
        
    def compare(self, obj_1, operator: CompareOperator, obj_2, is_obtained_vs_expected = True, raise_exception = True, redirect_to_equals=True):
        if operator == CompareOperator.Equal:
            return self.__equals(obj_1, obj_2, is_obtained_vs_expected, raise_exception)
        else:
            #TODO: Adapt compare method to cell value types as in __equals
            super().compare(obj_1, operator, obj_2, is_obtained_vs_expected, raise_exception, redirect_to_equals=redirect_to_equals)
    
    def __get_values_to_compare(self, cell_1, cell_2):
        compare_result = undefined_value
        
        if isinstance(cell_1, ValueTableCell):
            if cell_1.value_type == ValueTypes.NotApplicable:
                compare_result = True
            cell_1_value = cell_1.get_value()
        else:
            cell_1_value = cell_1.content
        
        if cell_2.value_type == ValueTypes.NotApplicable:
            compare_result = True
        cell_2_value = cell_2.get_value()
        
        return cell_1_value, cell_2_value, compare_result
        
    def __equals(self, cell_1, cell_2, is_obtained_vs_expected = True, raise_exception = True):
        res = False
        
        # Extract cell values to compare
        cell_1_value, cell_2_value, compare_result = self.__get_values_to_compare(cell_1, cell_2)
        if compare_result is not undefined_value:
            return compare_result
        
        cause = None
        try:
            # value_type = cell_2.value_type
            value_type = cell_2.content_type
            if value_type == ValueTypes.NotApplicable:
                res = True
            elif value_type == ValueTypes.Null:
                res = (cell_1_value is None)
            elif value_type in [ValueTypes.Boolean, ValueTypes.Integer]:
                res = (cell_1_value == cell_2_value)
            elif value_type == ValueTypes.Float:
                res = self.__get_float_comparator().equals(cell_1_value, cell_2_value, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
            elif value_type == ValueTypes.Generic:
                comparator = ObjectComparator()
                res = comparator.equals(cell_1_value, cell_2_value, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
            elif value_type == ValueTypes.Merged:
                return super().equals(cell_1, cell_2, is_obtained_vs_expected, raise_exception)
            elif ValueTypes.is_string(value_type):
                res = self._equals_string(cell_1, cell_2, is_obtained_vs_expected, raise_exception)
            elif value_type == ValueTypes.Symbol:
                res = self._equals_symbol(cell_1, cell_2, is_obtained_vs_expected, raise_exception)
            else:
                raise TechnicalException("Unmanaged value of type '{}'".format(value_type.name))
        except FunctionalException as exc:
            if not raise_exception:
                # Unexpected exception
                raise TechnicalException("Unexpected exception while raise_exception=False") from exc
            res = False
            cause = exc
        
        if not res and raise_exception:
            msg_list = [f"Cells are not equal",
                        f"    cell {self._get_name_1(is_obtained_vs_expected)}: {cell_1.represent()} (type: {Typing.get_object_class_fullname(cell_1_value)})"]
            msg_line = f"    cell {self._get_name_2(is_obtained_vs_expected)}: {cell_2.represent()} (scenario type: {value_type.name}"
            if cell_2_value != cell_2.content:
                msg_line += f" ; value: {cell_2_value} ; value type: {Typing.get_object_class_fullname(cell_2_value)}"
            msg_line += ")"
            msg_list.append(msg_line)
            msg = "\n".join(msg_list)
            if cause is not None:
                raise VerifyException(msg) from cause
            else:
                raise VerifyException(msg)
        return res
    
    def _equals_string(self, cell_1, cell_2, is_obtained_vs_expected = True, raise_exception = True):
        """
        Method used to compare cell_1 to string cell_2.
        First cell_1 content is verified to be a string, then cells contents are compared
        """
        
        content_1, content_2 = self._convert_inputs(cell_1, cell_2, is_obtained_vs_expected)
        
        # Verify cell_1 content type
        if not self.__type_comparator.equals(content_1, str, raise_exception=False):
            if raise_exception:
                raise VerifyException(f"{self._get_name_1(is_obtained_vs_expected).capitalize()} value is not a string: [{content_1}] (type: {Typing.get_object_class_fullname(content_1)})")
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Compare result is false since {self._get_name_1(is_obtained_vs_expected)} value is not a string: [{content_1}] (type: {Typing.get_object_class_fullname(content_1)})")
                return False
        
        # Verify cells content, without converting cell_1 content as string (cf property __string_comparator)
        return self.__string_comparator.equals(content_1, content_2, is_obtained_vs_expected, raise_exception)
        
    def _equals_symbol(self, cell_1, cell_symbol, is_obtained_vs_expected = True, raise_exception = True):
        res = True
        
        if isinstance(cell_1, ValueTableCell):
            cell_1_value = cell_1.value
        else:
            cell_1_value = cell_1.content
        
        symbol_value = cell_symbol.value
        if symbol_value is None:
            if cell_1_value is not None:
                if raise_exception:
                    raise VerifyException("Cell value is not None (cell: [{}])".format(cell_1))
                else:
                    res = False
        elif isinstance(symbol_value, str) and self.__get_text_verifier().is_to_interpret(symbol_value):
            # Case where symbol contains a verify function
            res = self.__get_text_verifier().verify(cell_1_value, symbol_value, raise_exception=raise_exception)
        else:
            comparator = ObjectComparator()
            res = comparator.equals(cell_1_value, symbol_value, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
        
        return res
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import queue
import logging
from holado.common.context.context import Context
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_ui.ui.drivers.ui_driver_info import UIDriverInfo
from holado_ui.ui.exceptions.focus_driver_exception import FocusDriverException
from holado_ui.ui.actors.actions import UIDriverAction, UIDriverCloser
from holado.common.context.session_context import SessionContext
from holado_ui.ui.gui.drivers.gui_driver import GUIDriver
from holado_ui_mss.ui.mss.mss_manager import MssManager
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class UIManager(Context):
    """ Class managing existing UIs (web browsers,...).
    The UI in top of stack is supposed active.
    """
    
    def __init__(self):
        super().__init__("UIManager")
        
        self.__focus_uid_stack = queue.LifoQueue()
        self.__driver_by_uid = {}
        
        self.__default_ui_lifetime_context = None
        self.__driver_initializers = []
        self.__driver_closers = []
        self.__default_driver_closer = None
        
        self.set_ui_lifetime_to_default()
        self.register_default_ui_driver_closer()
    
    def _delete_object(self):
        # Close all opened UIs
        self.close_uis()
        
        super()._delete_object()
    
    def has_focused_driver(self, uid=None):
        """
        @param uid Driver UID
        @return True if given driver (or any driver if uid is None) is registered in focus.
        """
        if uid is not None:
            return uid in self.__focus_uid_stack
        else:
            return not self.__focus_uid_stack.empty()
    
    def has_driver(self, uid):
        """
        @param uid Driver UID.
        @return True if driver of given UID is registered.
        """
        return uid in self.__driver_by_uid

    def get_driver_info(self, uid):
        """
        @param uid Driver UID
        @return UI driver info
        """
        if not self.has_driver(uid):
            raise TechnicalException(f"Driver '{uid}' is not registered")
        return self.__driver_by_uid[uid]
    
    def _get_driver_uids(self):
        return list(self.__driver_by_uid.keys())
    
    def _get_focused_driver_uids(self):
        return list(self.__focus_uid_stack.queue)

    @property
    def in_focus_driver(self):
        """
        @return Driver currently in focus.
        """
        self.verify_has_in_focus_driver()
        return self.in_focus_driver_info().driver

    @property
    def in_focus_driver_info(self):
        """
        @return Driver info of driver currently in focus.
        """
        return self.__driver_by_uid[self.__focus_uid_stack.queue[-1]]

    def pop_focused_driver(self, uid):
        """
        Unregister driver of given UID and return it.
        @param uid Driver UID.
        @return Driver instance.
        """
        return self.pop_focused_driver_info(uid).driver

    def pop_focused_driver_info(self, uid):
        """
        Unregister driver of given UID and return it as a driver info.
        @param uid Driver name.
        @return Driver info.
        """
        if Tools.do_log(logger, logging.DEBUG):
            msg_list = []
            msg_list.append(f"Popped driver '{uid}' (previous current: '{self.in_focus_driver_info().uid}'")
        
        # Search driver of given name
        if self.has_focused_driver(uid):
            res = self.get_driver_info(uid)
        else:
            raise FunctionalException(f"Driver '{uid}' is not registered in focus")

        # Remove it
        try:
            self.__focus_uid_stack.remove(uid)
        except Exception as exc:
            raise TechnicalException(f"Unable to remove driver '{uid}'") from exc

        if Tools.do_log(logger, logging.DEBUG):
            msg_list.append(f" ; new current: {self.in_focus_driver_info().uid if self.has_focused_driver() else 'None'})")
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("".join(msg_list))
        
        return res

    def pop_in_focus_driver(self):
        """
        Unregister driver currently in focus and return it.
        @return Driver instance.
        """
        return self.pop_in_focus_driver_info().driver

    def pop_in_focus_driver_info(self):
        """
        Unregister driver currently in focus and return it as a driver info.
        @return Driver info
        """
        if Tools.do_log(logger, logging.DEBUG):
            msg_list = []
            msg_list.append(f"Popped current driver (popped: '{self.in_focus_driver_info().uid}'")
        
        res = self.__driver_by_uid[self.__focus_uid_stack.pop()]

        if Tools.do_log(logger, logging.DEBUG):
            msg_list.append(f" ; new current: {self.in_focus_driver_info().uid if self.has_focused_driver() else 'None'})")
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("".join(msg_list))
        
        return res

    def push_driver(self, uid, driver, is_hidden):
        """
        Register a new driver and set it in focus if not hidden.
        @param uid Driver UID.
        @param driver Driver instance.
        @param is_hidden If true, driver is considered hidden and is not added in focused drivers
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Pushing{' hidden' if is_hidden else ''} driver '{uid}' (current driver: {self.in_focus_driver_info().uid if self.has_focused_driver() else 'None'})")
        
        self.__driver_by_uid[uid] = UIDriverInfo(uid, driver, is_hidden)
        if not is_hidden:
            self.__focus_uid_stack.put(uid)
        
        self.set_object(uid, driver, self.__default_ui_lifetime_context)

    def repush_driver_in_focus(self, uid):
        """
        Repush driver of given UID in focus
        @param uid Driver UID
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Repushing driver '{uid}' (current driver: {self.in_focus_driver_info().uid if self.has_focused_driver() else 'None'})")
        
        # Search driver of given name
        if not self.has_driver(uid):
            raise FunctionalException(f"Driver '{uid}' does not exist")
        driver_info = self.get_driver_info(uid)
        if driver_info.is_hidden:
            raise FunctionalException(f"Driver '{uid}' is hidden")

        # Remove it
        try:
            self.__focus_uid_stack.remove(uid)
        except Exception as exc:
            raise TechnicalException(f"Unable to remove driver '{uid}' before repushing it") from exc
        
        # Repush it
        self.__focus_uid_stack.put(uid)
        
    def switch_to_driver(self, uid):
        """
        Set focus on driver of given UID
        @param uid UID of already opened driver
        """
        # Repush driver
        self.repush_driver_in_focus(uid)

    
    def check_driver_is_in_focus(self, uid):
        """
        @param uid Drive UID
        @return True if driver of given UID is currently in focus
        """
        if self.has_focused_driver():
            return (self.in_focus_driver_info().uid == uid)
        else:
            return False
    
    def verify_driver_is_in_focus(self, uid):
        """
        If driver currently in focus is not of given UID, raise an exception.
        @param uid Driver UID
        """
        self.verify_has_in_focus_driver()
        if not self.check_driver_is_in_focus(uid):
            raise FocusDriverException(f"Driver '{uid}' is not in focus, it is driver '{self.in_focus_driver_info().uid}' (focused drivers: {self._represent_focused_driver_uids()})")
    
    def verify_has_in_focus_driver(self):
        """
        Raise an exception if no driver is currently registered.
        """
        if not self.has_focused_driver():
            raise FunctionalException("No driver is currently registered in focus")
    
    def register_ui_driver_initializer(self, driver_initializer:UIDriverAction):
        """
        Register a new UI driver initializer
        @param driver_initializer Driver initializer
        """
        self.__driver_initializers.append(driver_initializer)
    
    def initialize_ui_driver(self, uid, driver):
        """
        Apply all registered UI driver initializer
        @param uid Driver UID
        @param driver UI driver to initialize
        """
        for initializer in self.__driver_initializers:
            try:
                initializer.execute(uid)
            except Exception as exc:
                logger.error(f"Unable to initialize driver '{uid}' with initializer '{initializer.name}': {str(exc)}")
        
    def register_default_ui_driver_closer(self, driver_closer:UIDriverCloser=None):
        """
        Register default UI driver closer
        @param driver_closer Driver closer implementation
        """
        if driver_closer is None:
            driver_closer = UIDriverCloser("default", self)
        self.__default_driver_closer = driver_closer
    
    def register_ui_driver_closer(self, driver_closer:UIDriverCloser):
        """
        Register a new UI driver closer
        @param driver_closer Driver closer implementation
        """
        self.__driver_closers.append(driver_closer)
    
    def set_ui_lifetime_to_default(self):
        """
        Set UI lifetime to default context
        """
        self.set_ui_lifetime_to("Session")
    
    def set_ui_lifetime_to(self, lifetime_context):
        """
        Set default UI lifetime to given context
        @param lifetime_context Lifetime context
        """
        # Store context for UI created later
        self.__default_ui_lifetime_context = lifetime_context
        
        # Update lifetime of currently created UIs
        for uid in self._get_driver_uids():
            self.set_driver_lifetime_to(lifetime_context, uid)
    
    def set_driver_lifetime_to(self, lifetime_context, uid):
        """
        Set driver lifetime to given context
        @param lifetime_context Lifetime context
        @param uid Driver UID
        """
        if self.has_driver(uid):
            SessionContext.instance().lifetime_manager.update_object_lifetime(uid, self, lifetime_context)
        else:
            raise TechnicalException(f"No driver '{uid}' is registered")
    
    def close_uis(self):
        """
        Close all UIs
        """
        logger.info("Closing all UIs...")
        
        # Close other drivers
        for uid in self._get_driver_uids():
            self.close_driver(uid)
    
    def close_uis(self, lifetime_context):
        """
        Close UIs with lifetime of given context
        @param lifetime_context Lifetime context
        """
        logger.info(f"Closing UIs at {lifetime_context} context end...")
        
        # Close other drivers
        for uid in self._get_driver_uids():
            self.close_driver(lifetime_context, uid)
    
        logger.info(f"Finished closing UIs at {lifetime_context} context end.")
    
    def close_driver(self, uid, lifetime_context=None):
        """
        Close driver of given UID
        @param uid Driver UID
        @param lifetime_context Lifetime context
        @return True if a driver was closed
        """
        res = False
        
        if self.has_driver(uid):
            if lifetime_context is not None:
                driver_lifetime = SessionContext.instance().lifetime_manager.get_object_lifetime(uid, self)
                if driver_lifetime is not None and driver_lifetime == lifetime_context:
                    logger.info(f"Closing driver '{uid}' at {lifetime_context} context end")
                else:
                    return False
            else:
                logger.info(f"Closing driver '{uid}'")

            # Apply registered drivers
            for closer in self.__driver_closers:
                try:
                    res = closer.execute(uid)
                except Exception as exc:
                    logger.error(f"Unable to close driver '{uid}' with closer '{closer.name}': {str(exc)}")
                
                if res:
                    break
            
            # Else apply default closer
            if not res:
                self.__default_driver_closer.execute(uid)
        
        return res
    
    def make_screenshots_for_debug(self, destination_path, context_description):
        """
        Make screenshots for debug (monitors, current UI in focus).
        @param destination_path Destination path
        @param context_description Context description that will be inserted in file names
        """
        # Monitors
        MssManager.make_monitors_screenshots(destination_path, context_description)
        # PyAutoGUIManager.make_monitors_screenshots(destination_path, context_description)
        
        # Current UI in focus
        if self.has_focused_driver():
            driver = self.in_focus_driver()
            if isinstance(driver, GUIDriver):
                driver.make_screenshots_for_debug(destination_path, context_description)
    

    def _representDriverUIDs(self):
        return f"[{', '.join(self._get_driver_uids())}]"

    def _represent_focused_driver_uids(self):
        return f"[{', '.join(self._get_focused_driver_uids())}]"

    def _remove(self, name, remove_from_lifetime_manager):
        if self.has_focused_driver(name):
            self.pop_focused_driver_info(name)
        if self.has_driver(name):
            del self.__driver_by_uid[name]
        
        super.remove(name, remove_from_lifetime_manager)
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import socket
from holado_python.standard_library.socket.socket import SocketClient
import abc
from holado_multitask.multithreading.loopfunctionthreaded import LoopFunctionThreaded
import selectors
from holado.common.handlers.undefined import undefined_argument
import time
import ssl
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


##########################################################################
## Clients
##########################################################################


class NonBlockingSocketClient(SocketClient):
    """
    Base class for blocking socket client.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        self.__selector = selectors.DefaultSelector()
        
        # Note: __selector must be defined before, since Socket.__init__ can execute create_ipv4_socket
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)

    def _delete_object(self):
        if self.internal_socket:
            # Note: stop must be done before unregistering selector
            self.stop()
            self.__selector.unregister(self.internal_socket)
        
        super()._delete_object()
    
    def _register_socket(self, sock):
        events = selectors.EVENT_READ | selectors.EVENT_WRITE
        self.__selector.register(sock, events, data=self._data)
    
    def start(self, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        """Start client event loop.
        """
        kwargs = {'read_bufsize':read_bufsize, 'read_kwargs':read_kwargs, 'write_kwargs':write_kwargs}
        # Note: delay_between_run_sec must be None since _idle_sleep_delay is used to make sleeps
        thread = LoopFunctionThreaded(self._wait_and_process_events, kwargs=kwargs, register_thread=True, delay_before_run_sec=None, delay_between_run_sec=None)
        self._start_thread(thread)
    
    def _wait_and_process_events(self, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Wait and process events: begin")
        has_activity = False
        events = self.__selector.select(timeout=None)
        for key, mask in events:
            has_activity |= self._service_connection(key, mask, read_bufsize=read_bufsize, read_kwargs=read_kwargs, write_kwargs=write_kwargs)
        
        # Wait before next loop if no data was exchanged
        if not has_activity and self._idle_sleep_delay is not None:
            if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[{self.name}] Wait and process events: wait {self._idle_sleep_delay} s")
            time.sleep(self._idle_sleep_delay)
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Wait and process events: end")

    def _service_connection(self, key, mask, *, read_bufsize=1024, read_kwargs=None, write_kwargs=None):
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Process service connection ({key}, {mask}): begin")
        has_activity = False
        read_kwargs = read_kwargs if read_kwargs is not None else {}
        write_kwargs = write_kwargs if write_kwargs is not None else {}
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Process service connection ({key}, {mask}): receive data")
        sock = key.fileobj
        data = key.data
        if self.is_run_with_recv and mask & selectors.EVENT_READ:
            if self.is_with_ssl:
                # ssl doesn't suppôrt flags != 0
                flags = 0
            else:
                flags = read_kwargs.get('flags', 0)
            try:
                recv_data = sock.recv(read_bufsize, flags)
            except ssl.SSLWantReadError as exc:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[{self.name}] socket recv error: {exc}")
                recv_data = None
                has_activity = True     # Socket is not idle
            if recv_data:
                has_activity = True
                with self._data_lock:      # data is self._data
                    data.in_bytes += recv_data
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"[{self.name}] Received [{recv_data}] (type: {type(recv_data)} ; total: {len(data.in_bytes)})")
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Process service connection ({key}, {mask}): send data")
        if self.is_run_with_send and mask & selectors.EVENT_WRITE:
            with self._data_lock:      # data is self._data
                if data.out_bytes:
                    has_activity = True
                    try:
                        sent = sock.send(data.out_bytes)
                    except ssl.SSLWantWriteError as exc:
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"[{self.name}] socket send error: {exc}")
                        sent = 0
                        has_activity = True     # Socket is not idle
                    if sent > 0:
                        data.out_bytes = data.out_bytes[sent:]
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug(f"[{self.name}] Sent {sent} data (remaining to send: {len(data.out_bytes)})")
        
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[{self.name}] Process service connection ({key}, {mask}): end  => had activity: {has_activity}")
        return has_activity
    

class TCPNonBlockingSocketClient(NonBlockingSocketClient):
    """
    TCP socket client.
    """
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay, do_run_with_recv=do_run_with_recv, do_run_with_send=do_run_with_send)
    
    def create_ipv4_socket(self, host, port, **kwargs):
        socket_kwargs = self._new_ssl_context_if_required(**kwargs)
        
        if self.is_with_ssl:
            # sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # sock = Socket.create_connection((host, port), do_connect=False, **socket_kwargs)
            sock = socket.create_connection((host, port), **socket_kwargs)
            sock.setblocking(False)
            
            wrap_socket_kwargs = self._ssl_wrap_socket_kwargs
            do_handshake_on_connect = wrap_socket_kwargs.pop('do_handshake_on_connect', False)
            sock = self._ssl_context.wrap_socket(sock, server_hostname=host, do_handshake_on_connect=do_handshake_on_connect, **wrap_socket_kwargs)
            self._set_internal_socket(sock, is_ssl_handshake_done_on_connect=do_handshake_on_connect)
            
            # sock.connect((host, port))
        else:
            sock = socket.create_connection((host, port), **socket_kwargs)
            sock.setblocking(False)
            self._set_internal_socket(sock)
        
        # Register socket
        self._register_socket(sock)










#TODO: change use of this module by holado_helper.initialize_holado


import os
import sys


def insert_holado_helper_script():
    # Insert path to holado_helper
    holado_path = os.getenv('HOLADO_PATH')
    if holado_path is None:
        # If HolAdo sources are not cloned on this environment, use path within this installation
        from holado import get_holado_path
        holado_path = get_holado_path()
    sys.path.insert(0, os.path.join(holado_path, "src", "holado_helper", "script") )
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy
from holado.common.handlers.undefined import undefined_value


class EvaluateParameters(object):
    """ Evaluate parameters
    """
    
    __instance_nothing = None
    __instance_default = None
    __instance_default_without_raise = None
    
    def __init__(self, do_interpret=True, do_eval_variable=True, do_eval=True, \
                 raise_on_interpret_error=True, raise_on_variable_eval_error=True, raise_on_eval_error=True,
                 do_interpret_recursively=False, do_eval_variable_recursively=False,
                 result_type=undefined_value):
        """
        Constructor
        @param  
        """
        self.do_interpret = do_interpret
        self.do_interpret_recursively = do_interpret_recursively
        self.do_eval_variable = do_eval_variable
        self.do_eval_variable_recursively = do_eval_variable_recursively
        self.do_eval = do_eval
        self.raise_on_interpret_error = raise_on_interpret_error
        self.raise_on_variable_eval_error = raise_on_variable_eval_error
        self.raise_on_eval_error = raise_on_eval_error
        self.result_type = result_type
        self.result_is_str = result_type is str
    
    def __str__(self)->str:
        return f"{{interpret:({self.do_interpret},recursive:{self.do_interpret_recursively},raise:{self.raise_on_interpret_error}) ; eval variable:({self.do_eval_variable},recursive:{self.do_eval_variable_recursively},raise:{self.raise_on_variable_eval_error}) ; eval:({self.do_eval},raise:{self.raise_on_eval_error}) ; result type:{self.result_type}}}"
    
    def with_interpret(self, do_interpret):
        """
        Note: if wanted do_interpret is the same, self instance is returned, else a new one is returned
        @param do_interpret Wanted do_interpret
        @return Same parameters but with given do_interpret
        """
        if do_interpret == self.do_interpret:
            return self
        else:
            res = copy.deepcopy(self)
            res.do_interpret = do_interpret
            return res
        
    def with_interpret_recursively(self, do_interpret_recursively):
        """
        Note: if wanted do_interpret_recursively is the same, self instance is returned, else a new one is returned
        @param do_interpret Wanted do_interpret_recursively
        @return Same parameters but with given do_interpret_recursively
        """
        if do_interpret_recursively == self.do_interpret_recursively:
            return self
        else:
            res = copy.deepcopy(self)
            res.do_interpret_recursively = do_interpret_recursively
            return res
        
    def with_eval_variable(self, do_eval_variable):
        """
        Note: if wanted do_eval_variable is the same, self instance is returned, else a new one is returned
        @param do_eval_variable Wanted do_eval_variable
        @return Same parameters but with given do_eval_variable
        """
        if do_eval_variable == self.do_eval_variable:
            return self
        else:
            res = copy.deepcopy(self)
            res.do_eval_variable = do_eval_variable
            return res
        
    def with_eval_variable_recursively(self, do_eval_variable_recursively):
        """
        Note: if wanted do_eval_variable_recursively is the same, self instance is returned, else a new one is returned
        @param do_eval_variable_recursively Wanted do_eval_variable_recursively
        @return Same parameters but with given do_eval_variable_recursively
        """
        if do_eval_variable_recursively == self.do_eval_variable_recursively:
            return self
        else:
            res = copy.deepcopy(self)
            res.do_eval_variable_recursively = do_eval_variable_recursively
            return res
        
    def with_eval(self, do_eval):
        """
        Note: if wanted do_eval is the same, self instance is returned, else a new one is returned
        @param do_eval Wanted do_eval
        @return Same parameters but with given do_eval
        """
        if do_eval == self.do_eval:
            return self
        else:
            res = copy.deepcopy(self)
            res.do_eval = do_eval
            return res
        
    def with_raise(self, raise_exception=True):
        """
        Note: if raise booleans are same, self instance is returned, else a new one is returned
        @param raise_exception Whether raise exceptions or not
        @return Same parameters but with raise
        """
        return self.with_raise_on_interpret_error(raise_exception) \
            .with_raise_on_variable_eval_error(raise_exception) \
            .with_raise_on_eval_error(raise_exception)
        
    def with_raise_on_interpret_error(self, raise_on_interpret_error=True):
        """
        Note: if raise_on_interpret_error booleans are same, self instance is returned, else a new one is returned
        @param raise_on_interpret_error Whether raise exceptions or not on interpret error
        @return Same parameters but with raise_on_interpret_error
        """
        if self.raise_on_interpret_error == raise_on_interpret_error:
            return self
        else:
            res = copy.deepcopy(self)
            res.raise_on_interpret_error = raise_on_interpret_error
            return res
        
    def with_raise_on_variable_eval_error(self, raise_on_variable_eval_error=True):
        """
        Note: if raise_on_variable_eval_error booleans are same, self instance is returned, else a new one is returned
        @param raise_on_variable_eval_error Whether raise exceptions or not on variable_eval error
        @return Same parameters but with raise_on_variable_eval_error
        """
        if self.raise_on_variable_eval_error == raise_on_variable_eval_error:
            return self
        else:
            res = copy.deepcopy(self)
            res.raise_on_variable_eval_error = raise_on_variable_eval_error
            return res
        
    def with_raise_on_eval_error(self, raise_on_eval_error=True):
        """
        Note: if raise_on_eval_error booleans are same, self instance is returned, else a new one is returned
        @param raise_on_eval_error Whether raise exceptions or not on eval error
        @return Same parameters but with raise_on_eval_error
        """
        if self.raise_on_eval_error == raise_on_eval_error:
            return self
        else:
            res = copy.deepcopy(self)
            res.raise_on_eval_error = raise_on_eval_error
            return res
        
    def without_raise(self):
        """
        Note: if raise booleans are already False, self instance is returned, else a new one is returned
        @return Same parameters but without raise
        """
        return self.with_raise(False)
        
    def with_result_type(self, result_type=undefined_value):
        """
        Note: if result_type types are same, self instance is returned, else a new one is returned
        @param result_type Evaluation result type
        @return Same parameters but with result_type
        """
        if self.result_type == result_type:
            return self
        else:
            res = copy.deepcopy(self)
            res.result_type = result_type
            res.result_is_str = result_type is str
            return res
    
    @staticmethod
    def default(raise_exception=True):
        """
        @param raise_exception Whether raise exceptions or not
        @return Default evaluate parameters with given raiseability
        """
        if raise_exception:
            return EvaluateParameters.default_with_raise()
        else:
            return EvaluateParameters.default_without_raise()
    
    @staticmethod
    def default_with_raise():
        """
        @return Default evaluate parameters with raises
        """
        if EvaluateParameters.__instance_default is None:
            EvaluateParameters.__instance_default = EvaluateParameters(True, True, True, True, True, True, undefined_value)
        return EvaluateParameters.__instance_default
    
    @staticmethod
    def default_without_raise():
        """
        @return Default evaluate parameters without any raise
        """
        if EvaluateParameters.__instance_default_without_raise is None:
            EvaluateParameters.__instance_default_without_raise = EvaluateParameters(True, True, True, False, False, False, undefined_value)
        return EvaluateParameters.__instance_default_without_raise
    
    @staticmethod
    def default_without_eval(raise_exception=True):
        """
        @param raise_exception Whether raise exceptions or not
        @return Default evaluate parameters with given raiseability and without eval
        """
        return EvaluateParameters.default(raise_exception).with_eval(False)
    
    @staticmethod
    def nothing():
        """
        @return Default evaluate parameters without any raise
        """
        if EvaluateParameters.__instance_nothing is None:
            EvaluateParameters.__instance_nothing = EvaluateParameters(False, False, False, False, False, False, undefined_value)
        return EvaluateParameters.__instance_nothing





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_ui.ui.gui.drivers.gui_driver import GUIDriver
from urllib.parse import unquote
from holado_ui_selenium.ui.gui.selenium.handlers.enums import BrowserTypes
from holado_core.common.handlers.redo import Redo
from holado_ui.ui.gui.windows.gui_window_info import GUIWindowInfo
from selenium.common.exceptions import NoSuchWindowException
from holado.common.handlers.enums import ObjectStates
from holado_core.common.exceptions.timeout_exception import TimeoutException
from holado_ui_selenium.ui.gui.selenium.tools.selenium_path_manager import SeleniumPathManager
from holado_ui_selenium.ui.gui.selenium.drivers.selenium_internal_api import SeleniumInternalAPI
from holado_ui_selenium.ui.gui.selenium.drivers.web_driver_manager import WebDriverManager
import urllib
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class SeleniumDriver(GUIDriver):
    """ Selenium driver.
    """
    
    def __init__(self, browser_type, path_manager, name=None):
        super().__init__(name or "SeleniumDriver")
        
        self.__path_manager = SeleniumPathManager(path_manager)
        self.__web_driver_manager = WebDriverManager(browser_type, self.__path_manager)
        self.__current_host = None
    
    def _initialize_internal_api(self):
        return SeleniumInternalAPI(self)
    
    @property
    def current_host(self):
        """
        @return Current host
        """
        return self.__current_host
    
    @current_host.setter
    def current_host(self, host):
        """
        Update current host
        @param host New host
        """
        if host != self.__current_host:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"current host has changed ('{self.current_host}' -> '{host}')")
            self.__current_host = host
    
    @property
    def browser_type(self):
        """
        @return the browser type
        """
        return self.__web_driver_manager.browser_type
    
    @property
    def _selenium_path_manager(self):
        return self.__path_manager
    
    @property
    def _path_manager(self):
        return self.__path_manager.path_manager
    
    @property
    def current_url(self, ):
        """
        @return current URL (it is unescaped).
        """
        # Get URL from selenium
        res = self.internal_driver.current_url
        # Decode URL (for percent-encoded data like %20)
        try:
            res = unquote(res, encoding='utf-8', errors='strict')
        except UnicodeError as exc:
            raise TechnicalException(str(exc)) from exc
        return res
        
    def open_site(self, host):
        """
        Open a portal given a specific host
        @param host Name of remote host
        """
        self.current_host = host
        self.open_current_host()
    
    def open_new_window(self):
        """
        Open a new page
        """
        if not self.is_open:
            self.open()
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Opening new window...")
            
            if self.has_current_window():
                self.internal_api.open_new_window()
            else:
                self.internal_api.focus()      # When entering in this case, close of last page was simulated, and the browser window has possibly lost focus
            
            self.check_new_browser_page_is_open()
            
    def open_current_host(self):
        """
        Open current host
        """
        # Build URL
        url = "http://" + self.current_host
        
        # Open URL
        # Note: use open_url rather than openSite in order to use same proxy settings
        self.open_url(url)
    
    def refresh(self):
        """
        Refresh current Html page
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Refreshing current page...")
        self.internal_api.refresh_browser()
        self.internal_api.wait_until_window_is_loaded()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Refreshed current page")
        
    def open(self):
        """
        Open a WebDriver for configured browser.
        """
        self.__open_driver()
        
    def __open_driver(self):
        """
        Open a WebDriver for configured browser.
        """
        # Create web driver
        self.internal_driver = self.__web_driver_manager.create_web_driver()
        
        self.internal_api.maximize()
        
      
    def _close_driver(self):
        super()._close_driver()
        self.__close_browser(self.__browser_type)
        
    def __close_browser(self, browser_type):
        if self.has_internal_driver:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("close web driver")
                
            # Close browser
#          if (CommonSystem.getOSType() == OSType.Windows):
#              #WORKAROUND: scenario stops and displays an error window saying plugin container has stopped
#              driver.quit()
#              Tools.sleep(5000)
#          }    

            # Quit web driver
#          if (browser_type == BrowserTypes.Firefox):
#            if (browser_type == BrowserTypes.Chrome):
#                try:
#                    driver.quit()
#                except Exception e):
#                    logger.warn("exception catched when quiting web driver: {}", e.getMessage())
#                #                        setInternalDriver(None)
            
            # Remove opened pages information
            self._clear_window_info()
            
    def open_url(self, url):
        """
        Open given URL in current active browser.
        @param url URL to open.
        """
        # Open new browser if needed
        if not self.is_open:
            self.open()
            self.check_new_browser_page_is_open()
                
        # Go to url
        self.navigate_to_url(url)

        # Update current host
        current_url = self.current_url
        current_host = self.__get_host_from_url(current_url)
        if current_host is None:
            raise FunctionalException(f"Url '{current_url}' does not match expected formats")
        else:
            self.current_host = current_host
        
        # Update managed handle and page
        page_handle = self.internal_driver.current_window_handle
        if current_host is not None:
            page_name = self._get_page_name_from_url(current_url)
            selenium_window = self._new_window_for_name(page_name)
        else:
            page_name = None
            selenium_window = self.new_default_window()
            selenium_window.initialize(self)
        window_info = GUIWindowInfo(selenium_window, page_name, page_handle)
        self._add_window_info(page_handle, window_info)
        if not self.has_current_window or self._get_current_window_id() != page_handle:
            self._push_current_window_id(page_handle)
        
    def __navigate_to_url(self, url):
        self.internal_driver.get(url)
        
    def check_browser_displays_page(self, page_name, host=None, timeout_seconds=None):
        """
        Check that browser displays given page until given timeout, and update internal current page.
        @param page_name Page name
        @param host Host name
        @param timeout_seconds Timeout (if None, default timeout is used)
        """
        if host is None:
            host = "https?://" + self.current_host
            
        self.internal_api.wait_until_window_is_loaded()
        
        # Check if current url contains expected host and page name
        class PageRedo(Redo):
            def __init__(self, driver):
                super().__init__(f"check browser displays page '{page_name}'")
                self.__driver = driver
                
            def _process(self):
                res = None
                
                current_page_name = self.__driver._get_page_name_from_url(self.__driver.current_url, host)
                if current_page_name is not None:
                    if self.text_verifier.check(page_name, current_page_name):
                        res = current_page_name
                else:
                    logger.warning(f"browser page '{self.current_url}' doesn't match expected page '{page_name}'")
                
                return res
            
        redo = PageRedo(self)
        redo.redo_while_none()
        if timeout_seconds is not None:
            redo.with_timeout(timeout_seconds)
        redo.with_nb_successive_failure_before_error(3)
        try:
            current_page_name = redo.execute()
        except TimeoutException:
            raise FunctionalException(f"Page '{page_name}' is not displayed (current: '{self.current_url}')")
        
        self.update_current_window(current_page_name, self.internal_driver.current_window_handle)
        
    def is_browser_displaying_page(self, page_name, host=None):
        """
        @param page_name Page name
        @param host Host name
        @return True if browser displays given page
        """
        if host is None:
            host = "http://" + self.current_host
            
        self._verify_is_open()
        self.internal_api.wait_until_window_is_loaded()
        
        # Return if current url contains expected host and page name
        current_page_name = self._get_page_name_from_url(self.current_url, host)
        if current_page_name is not None:
            res = self.text_verifier.check(page_name, current_page_name)
        else:
            res = False
        
        return res
        
    def check_new_browser_page_is_open(self):
        """
        Checks and register a new opened browser.
        If more than one new browsers are detected, an exception is thrown 
        """
        class PageRedo(Redo):
            def __init__(self, driver):
                super().__init__(f"check new browser is open")
                self.__driver = driver
                
            def _process(self):
                res = None
                
                handles = self.internal_api.window_handles
                
                # Verify that one and only one new browser exists
                if len(handles) > self._nb_windows() + 1:
                    raise FunctionalException(f"More than one ({len(handles) - self._nb_windows()} new windows exist. Unable to determine which one is the last one.")
                elif len(handles) == self._nb_windows():
                    raise NoSuchWindowException("No new browser is open")
                
                # Find and register new browser handle
                for handle in handles:
                    # If new browser exists
                    if not self._has_window_info(handle):
                        # Move to new browser
                        self.switch_to_window(handle)
                        self._push_current_window_id(handle)
                        res = handle
                        break
                
                return res
            
        redo = PageRedo(self)
        redo.redo_while_none()
        redo.ignoring(NoSuchWindowException)
        redo.with_nb_successive_failure_before_error(3)
        new_handle = redo.execute()
        
        # Check new window is found
        if new_handle is None:
            raise FunctionalException("Unable to find new browser")
           
        # Update current host
        url = self.current_url
        self.current_host = self.__get_host_from_url(url)
        
        # update managed handle and page
        if self.current_host is not None:
            page_name = self._get_page_name_from_url(url)
            selenium_window = self._new_window_for_name(page_name)
        else:
            page_name = None
            selenium_window = self.new_default_window()
            selenium_window.initialize(self)
        self._add_window_info(new_handle, GUIWindowInfo(selenium_window, page_name, new_handle))
        
    def __get_host_from_url(self, str_url):
        split_result = urllib.parse.urlsplit(str_url)
        res_split = urllib.parse.SplitResult(scheme=split_result.scheme, netloc=split_result.netloc, path='', query='', fragment='')
        return res_split.geturl()
        
    def _get_page_name_from_url(self, url, host=None):
        if host is None:
            if self.current_host is None:
                raise TechnicalException("Hostname is not defined")
            host = self.current_host
            
        # Verify host
        url_host = self.__get_host_from_url(url)
        if url_host != host:
            raise FunctionalException(f"Unexpected host in url '{url}' (obtained: '{url_host}'  expected: '{host}')")
        
        # Get page name
        split_result = urllib.parse.urlsplit(url)
        return split_result.path
    
    def _close_current_window(self):
        """
        Close current window.
        If is under closing, close of last window, else simulate a close to keep browser and web driver connected
        """
        is_closing = (self.object_state == ObjectStates.Closing)
        self.__close_current_window(is_closing)
    
    """
    Close current window.
    @param close_last_window If True close last window, else simulate a close to keep browser and web driver connected
    """
    def __close_current_window(self, close_last_window):
        # Disactivate JS onbeforeunload
        self.internal_api.disactivate_js_onbeforeunload()

        # Close current browser page
        if self._nb_windows() > 1 or close_last_window:
            self.internal_driver.close()
        else:
            # Change page to default
            # Note: currently, pages that doesn't start with http are not managed
#          self.open_url("about:blank")
            self.open_url("https://www.lilo.org/fr/")
        
        # Leave handle internally
        handle = self._get_current_window_id()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Leaving window {handle}")
        
        # Close browser or move to previous page
        if self._nb_windows() == 0:
            if close_last_window:
                # No window exists anymore, close browser 
                if self.__browser_type == BrowserTypes.Firefox:
                    handles = self.internal_api.window_handles
                    if len(handles) == 0 or len(handles) == 1 and handle in handles:
                        self.close()
                else:
                    self.close()
        else:
            #  Move to previous page
            self.set_focus_on_current_page()
            self.current_host = self.__get_host_from_url(self.current_url)
        
    def set_focus_on_current_page(self):
        """
        Set focus on current registered page.
        It is used internally by pages when a popup is closed, or when a page is closed. 
        """
        self.__switch_to_window(self._get_current_window_id())
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"return into window {self._get_current_window_id()}")
        
    def __switch_to_window(self, handle):
        try:
            self.internal_driver.swith_to.window(handle)
        except NoSuchWindowException as exc:
            msg_list = [str(exc)]
            msg_list.append(f"    Known by framework: {self._get_all_window_ids()}")
            msg_list.append(f"    Known by selenium: {self.internal_api.window_handles}")
            raise NoSuchWindowException("\n".join(msg_list)) from exc
            
    def make_screenshots_for_debug(self, destination_path, context_description):
        # Screenshot of current window
        self.internal_api.make_screenshot(destination_path, context_description)
        
        # Backup of current window page source
        self.internal_api.makePageSourceBackup(destination_path, context_description)
    
    
    
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.finders.tools.find_context import FindContext,\
    ContainerFindContext, ListContainersFindContext
from holado_core.common.finders.tools.find_parameters import FindParameters
import logging
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.handlers.element_holder import ElementHolder
from holado_core.common.finders.tools.find_builder import FindBuilder
from holado_core.common.finders.tools.find_updater import FindUpdater
from holado_core.common.finders.tools.enums import FindType
from holado_core.common.exceptions.element_exception import NoSuchElementException,\
    TooManyElementsException
from holado_core.common.finders.tools.finder_info import FinderInfo
import abc

logger = logging.getLogger(__name__)


class Finder(object):
    """ Generic finder class.
    
    Any inheriting class should override the appropriate methods.
    Usually, methods _find_all_in_XXX are overridden, except _find_all_in_candidates.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, description=None):
        """
        @param description Description.
        """
        self.__description = description
        self.__inspector = None
        # self.__info = None
        self.__find_builder = None
        self.__find_updater = None
    
    @property
    def has_info(self):
        """
        @return If has finder information
        """
        return self.__info is not None
    
    @property
    def info(self):
        """
        @return Finder information
        """
        if self.__info == None:
            self.__info = FinderInfo()
        return self.__info
    
    @property
    def inspector(self):
        return self.__inspector
    
    @inspector.setter
    def inspector(self, inspector):
        """
        @param inspector Inspector that has created self finder
        """
        self.__inspector = inspector
    
    @property
    def find_builder(self):
        """
        Priors find builder from inspector
        @return Find builder
        """
        if self.__inspector != None:
            return self.__inspector.find_builder
        else:
            if self.__find_builder == None:
                self.__find_builder = FindBuilder()
            return self.__find_builder

    @find_builder.setter
    def find_builder(self, find_builder):
        """
        @param find_builder Find builder to use
        """
        self.__find_builder = find_builder
    
    @property
    def has_find_updater(self):
        """
        @return If has find updater
        """
        return self.__find_updater is not None
    
    @property
    def find_updater(self):
        """
        @return Find updater
        """
        if self.__find_updater == None:
            self.__find_updater = FindUpdater()
        return self.__find_updater
    
    @property
    def find_type(self):
        """
        @return Specific find method type. 
        """
        return self.find_updater.find_type
    
    @find_type.setter
    def find_type(self, find_type):
        """
        Set a specific find method type.
        @param find_type Find method type.
        """
        self.find_updater.find_type = find_type
    
    @property
    def finder_description(self):
        return f"{self.__class__.__name__}({self.element_description})"
    
    @property
    def element_description(self):
        """
        @return Description of the element(s) to find.
        """
        return self.__description
    
    
    def update_context(self, find_context):
        """
        @param find_context Context
        @param find_parameters Parameters
        @return Updated context 
        """
        if self.has_find_updater:
            return find_context.update(self.find_updater)
        else:
            return find_context
    
    def update_parameters(self, find_parameters):
        """
        @param find_parameters Parameters
        @return Updated parameters 
        """
        if self.has_find_updater:
            return find_parameters.update(self.find_updater)
        else:
            return find_parameters
    
    def find_in(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        return self.find(FindType.In, container, candidates, find_context, find_parameters)
    
    def find(self, find_type=None, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        Find element. 
        @param find_type Find type
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Element found.
        """
        if find_type is None:
            find_type = self.find_type
        find_context = self.find_builder.context(find_context, find_type, container, candidates)
        find_parameters = self.find_builder.parameters(find_parameters)

        # Process finders
        candidates = self._find_all(find_context=find_context, find_parameters=find_parameters)
        
        # Get element from candidates
        return self._get_element_from_list(candidates, find_parameters)
    
    def find_all_in(self, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        return self.find_all(FindType.In, container, candidates, find_context, find_parameters)
    
    def find_all(self, find_type=None, container=None, candidates=None, find_context=None, find_parameters=None):
        """
        Find all elements. 
        @param find_type Find type
        @param container Container from which make search.
        @param candidates Candidate containers.
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if find_type is None:
            find_type = self.find_type
        find_context = self.find_builder.context(find_context, find_type, container, candidates)
        find_parameters = self.find_builder.parameters(find_parameters)
        
        return self._find_all(find_context, find_parameters)
    
    def _find_all(self, find_context:FindContext, find_parameters:FindParameters):
        """
        Find all elements. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if isinstance(find_context, ContainerFindContext):
            return self._find_all_container(self.update_context(find_context), self.update_parameters(find_parameters))
        elif isinstance(find_context, ListContainersFindContext):
            return self._find_all_candidates(self.update_context(find_context), self.update_parameters(find_parameters))
        else:
            raise TechnicalException(f"[{self.finder_description}] Unmanaged find context '{find_context}'")
    
    def _find_all_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        """
        Find all elements for given container. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if self.find_type == FindType.Custom:
            # Unexpected case, self method should be overridden
            raise NotImplementedError(f"[{self.finder_description}] {self}")
        
        if find_context.find_type == FindType.In:
            return self._find_all_in_container(find_context, find_parameters)
        else:
            raise TechnicalException(f"[{self.finder_description}] Unmanaged find type '{find_context.find_type.name}'")
    
    def _find_all_candidates(self, find_context:ListContainersFindContext, find_parameters:FindParameters):
        """
        Find all elements for given candidates. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if self.find_type == FindType.Custom:
            # Default implementation is to iterate on containers
            return self._find_all_by_iteration(find_context, find_parameters)
        
        if find_context.find_type == FindType.In:
            return self._find_all_in_candidates(find_context, find_parameters)
        else:
            raise TechnicalException(f"[{self.finder_description}] Unmanaged find type '{find_context.find_type.name}'")
        
    
    def _find_all_by_iteration(self, find_context:ListContainersFindContext, find_parameters:FindParameters):
        """
        Find all elements by iterating on each container. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        res = []
        exc_intermediate = None
        
        prefix_logs = f"{self._get_indent_string_level(find_parameters)}[Finder({self.element_description}).findAll{'' if self.find_type is None else self.find_type.name}]"
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"{prefix_logs} -> begin with {find_context.nb_containers} candidates")
        
        # Iterate on each container context
        for index, container_find_context in enumerate(find_context):
            container_context = self.update_context(container_find_context)
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            
                logger.trace(f"{prefix_logs}     -> candidate {index}: {container_context.input_description}")
            
            try:
                cand_res = self._find_all(container_context, find_parameters)
                res.extend(cand_res)
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{prefix_logs}       -> candidate {index}   => {len(cand_res)} elements found")
            except NoSuchElementException as exc:
                if exc_intermediate is None:
                    exc_intermediate = exc
            
        if len(res) == 0 and exc_intermediate is not None:
            raise exc_intermediate

        if len(res) > 0:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{prefix_logs} -> return 0 candidate")
        else:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{prefix_logs} -> return {len(res)} candidates:\n{self._represent_candidates_output(res, 4)}")
        return res
        
    
    def _find_all_in_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        """
        Find all elements in given container. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if find_context.find_type == FindType.In:
            # Unexpected case, self method should be overridden
            raise NotImplementedError(f"[{self.finder_description}] {self}")
        else:
            return self._find_all_container(find_context.with_find_type(FindType.In), find_parameters)
    
    def _find_all_in_candidates(self, find_context:ListContainersFindContext, find_parameters:FindParameters):
        """
        Find all elements in given candidates. 
        @param find_context Find context
        @param find_parameters Find parameters
        @return Elements found.
        """
        if find_context.find_type == FindType.In:
            # Default implementation is to iterate on containers
            return self._find_all_by_iteration(find_context, find_parameters)
        else:
            return self._find_all_candidates(find_context.with_find_type(FindType.In), find_parameters)
    
    def _get_element_from_list(self, candidates, find_parameters):
        if find_parameters.nth_element is not None:
            return self._get_nth_element_from_list(candidates, find_parameters)
        
        res = None
        
        # Analyze search result
        if len(candidates) == 1:
            res = candidates[0]
        elif len(candidates) > 1:
            raise TooManyElementsException(f"More than one ({len(candidates)}) {self.element_description} were found:\n{self.represent_candidate_output(candidates, 4)}")
        elif find_parameters.raise_no_such_element:
            raise NoSuchElementException(f"Unable to find {self.element_description}")
           
        return res
    
    def _get_nth_element_from_list(self, candidates, find_parameters):
        res = None
        
        if find_parameters.nth_element is not None and find_parameters.nth_element < 1:
            raise TechnicalException(f"Nth ({find_parameters.nth_element}) must be positive (>= 1)")
            
        # Analyze search result
        if len(candidates) < 1:
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException(f"Unable to find {self.element_description}: no candidates were found")
        elif len(candidates) < find_parameters.nth_element:
            msg = f"Unable to find {self.element_description}: Nth ({find_parameters.nth_element}) is greater than number of candidates ({len(candidates)})"
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException(msg)
            else:
                logger.warning(self._get_indent_string_level(find_parameters) + msg)
        else:
            res = candidates[find_parameters.nth_element - 1]
           
        return res
    

    def verify_find_type_is_defined(self, error_message):
        """
        @param error_message Error message to raise is find type is not defined.
        """
        if self.find_type == FindType.Undefined:
            raise TechnicalException(f"[{self.element_description}] {error_message}")
    
    def is_valid_input(self, container, find_context, find_parameters):
        """
        @param container Container to validate
        @param find_context Find context
        @param find_parameters Find parameters
        @return True if container is valid according this finder
        """
        raise NotImplementedError(f"Unimplemented is_valid_input since inconsistent for [{self.finder_description}]")
        
    def _is_valid_input_in(self, finders, container, find_context, find_parameters):
        """
        Internal use, to check if container is valid in any finder of given list.
        @param finders Finder list
        @param container Container to validate
        @param find_parameters Find parameters
        @return True if container is valid for any finder
        """
        for finder in finders:
            try:
                if finder is not None:
                    if finder.is_valid_input(container, find_context, find_parameters):
                        return True
            except NotImplementedError:
                # continue
                pass
            
        return False
    
    def is_valid_output(self, element, find_context, find_parameters):
        """
        Return if element is valid according this finder.
        @param element Element to validate
        @param find_context Find context
        @param find_parameters Find parameters
        @return True if element is valid according this finder
        """
        raise NotImplementedError(f"Unimplemented is_valid_output since inconsistent for [{self.finder_description}]")
        
    def _is_valid_output_in(self, finders, element, find_context, find_parameters):
        """
        Internal use, to check if element is valid in any finder of given list.
        @param finders Finder list
        @param element Element to validate
        @param find_parameters Find parameters
        @return True if element is valid for any finder
        """
        for finder in finders:
            try:
                if finder is not None:
                    if finder.is_valid_output(element, find_context, find_parameters):
                        return True
            except NotImplementedError:
                # continue
                pass
            
        return False
    
    def _represent_candidates_output(self, candidates, indent):
        res_list = []
        for index, cand in candidates:
            res_list.append(f"{index}: {self._represent_candidate_output(cand)}")
        return Tools.indent_string(indent, "\n".join(res_list))

    def _represent_candidate_output(self, candidate):
        raise NotImplementedError(f"Missing implementation in {self}")
    
    def _get_indent_string_level(self, find_parameters: FindParameters=None, level=None):
        if level is not None:
            return Tools.get_indent_string(level * 4)
        elif find_parameters is not None:
            return self._get_indent_string_level(level=find_parameters.find_level)
        else:
            raise TechnicalException("level or find_parameters must be defined")

    def _add_candidates(self, res, candidates):
        for candidate in candidates:
            self._add_candidate(res, candidate)
    
    def _add_candidate(self, res, candidate):
        if not candidate in res:
            res.append(candidate)
    

    def _build_result(self, res, find_context, find_parameters):
        # Manage generic finders that don't heritate from ElementFinder, like ElseFinder
        if isinstance(res, ElementHolder):
            self._build_result_element(res, find_context, find_parameters)
        return res
    
    def _build_result_element(self, res, find_context, find_parameters):
        res.update_find_info(self, find_context, find_parameters)
        return res
    
    def _build_result_list(self, res, find_context, find_parameters):
        for el in res:
            self._build_result(el, find_context, find_parameters)
        return res
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import os
import logging
from holado_report.report.execution_historic import ExecutionHistoric
from holado.common.context.session_context import SessionContext
from holado_core.common.tools.tools import Tools
import threading

logger = logging.getLogger(__name__)



class BaseReport():
    def __init__(self):
        self.__report_path = None
        self.__execution_historic = None
        self.__report_builders = []
        
        self.__children_reports_lock = threading.RLock()
        self.__children_reports_by_prefix = {}
        
    def initialize(self):
        """
        Implement this method to initialize report.
        """
        pass
        
    def initialize_reports(self):
        """
        Implement this method to configure and initialize execution historic and report builders.
        """
        pass
    
    @property
    def has_report_path(self):
        return self.__report_path is not None
    
    @property
    def report_path(self):
        return self.__report_path
    
    @report_path.setter
    def report_path(self, report_path):
        self.__report_path = report_path
    
    @property
    def execution_historic(self):
        return self.__execution_historic
    
    @property
    def report_builders(self):
        return self.__report_builders
    
    def children_reports(self, prefix):
        with self.__children_reports_lock:
            if prefix in self.__children_reports_by_prefix:
                return self.__children_reports_by_prefix[prefix]
            else:
                return None
        
    def get_path(self, *args):
        return os.path.join(self.__report_path, *args)
        
    def set_execution_historic(self):
        self.__execution_historic = ExecutionHistoric()
        
    def add_report_builder(self, report_builder):
        self.__report_builders.append(report_builder)
        
    def build_reports(self):
        for rb in self.__report_builders:
            rb.build()

    def add_child_report(self, report, report_prefix, report_name, report_details=None):
        with self.__children_reports_lock:
            if report_prefix not in self.__children_reports_by_prefix:
                self.__children_reports_by_prefix[report_prefix] = []
                
            # Define child report path
            if self.has_report_path:
                report_dir_name = f"{report_prefix}_{len(self.__children_reports_by_prefix[report_prefix])+1:03d}"
                report.report_path = self.get_path(f"{report_prefix}s".capitalize(), report_dir_name)
                SessionContext.instance().path_manager.makedirs(report.report_path, is_directory=True)
            
            # Add child report
            self.__children_reports_by_prefix[report_prefix].append((report_name, report))
            
            # Update text file describing children reports
            if self.has_report_path:
                with open(self.get_path(f"{report_prefix}s".capitalize(), f"{report_prefix}_names.txt"), "at") as f:
                    f.write(f"{report_dir_name}: {report_name}\n")
                    if report_details:
                        f.write(Tools.indent_string(len(report_dir_name)+2, report_details) + "\n")
            

    def before_all(self):
        if self.__execution_historic:
            self.__execution_historic.before_all()
        for rb in self.__report_builders:
            rb.before_all()
    
    def before_feature(self, feature_context, feature, feature_report=None):
        if self.__execution_historic:
            self.__execution_historic.before_feature(feature_context, feature, feature_report)
        for rb in self.__report_builders:
            rb.before_feature(feature_context, feature, feature_report)
    
    def before_scenario(self, scenario_context, scenario, scenario_report=None):
        if self.__execution_historic:
            self.__execution_historic.before_scenario(scenario_context, scenario, scenario_report)
        for rb in self.__report_builders:
            rb.before_scenario(scenario_context, scenario, scenario_report)
    
    def before_step(self, step_context, step, step_level):
        if self.__execution_historic:
            self.__execution_historic.before_step(step_context, step, step_level)
        for rb in self.__report_builders:
            rb.before_step(step_context, step, step_level)
    
    def after_step(self, step_context, step, step_level):
        if self.__execution_historic:
            self.__execution_historic.after_step(step_context, step, step_level)
        for rb in self.__report_builders:
            rb.after_step(step_context, step, step_level)
    
    def after_scenario(self, scenario, scenario_report=None):
        if self.__execution_historic:
            self.__execution_historic.after_scenario(scenario, scenario_report)
        for rb in self.__report_builders:
            rb.after_scenario(scenario, scenario_report)
    
    def after_feature(self, feature, feature_report=None):
        if self.__execution_historic:
            self.__execution_historic.after_feature(feature, feature_report)
        for rb in self.__report_builders:
            rb.after_feature(feature, feature_report)
    
    def after_all(self, build_reports=True):
        if self.__execution_historic:
            self.__execution_historic.after_all()
        for rb in self.__report_builders:
            rb.after_all()
        
        if build_reports:
            self.build_reports()
        
    def release_resources(self):
        if self.__execution_historic:
            self.__execution_historic = None





from django.contrib.auth.models import User, Group
from rest_framework import serializers


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = ['url', 'username', 'email', 'groups']


class GroupSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Group
        fields = ['url', 'name']
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.drivers.internal_api import InternalAPI
from holado_core.common.finders.tools.find_context import ContainerFindContext,\
    ListContainersFindContext
from holado_core.common.exceptions.element_exception import TooManyElementsException,\
    NoSuchElementException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class ElementInternalAPI(InternalAPI):
    """ Base class for element internal API.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, driver):
        super().__init__(driver)
    
    def find_element(self, criteria, find_context, find_parameters):
        """
        Find element with given criteria
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
            return self.find_element_by_container(criteria, find_context.container, find_context, find_parameters)
        elif isinstance(find_context, ListContainersFindContext) and find_context.nb_containers > 0:
            return self.find_element_by_candidates(criteria, find_context.containers, find_context, find_parameters)
        else:
            raise TechnicalException("Find context has no container")
    
    
    def find_element_by_container(self, criteria, container, find_context, find_parameters):
        """
        Find element with given criteria and container
        
        A default implementation use find_elements_by_container.
        It is usually recommended to override it with a dedicated implementation.
        @param criteria Criteria
        @param container Container
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        # Manage no candidates
        if container is None:
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException("No container")
            else:
                return None
        
        found_elements = self.find_elements_by_container(criteria, container, find_context, find_parameters)
        return self._get_element_from_list(found_elements, 1, criteria, find_context, find_parameters)
    
    def find_element_by_candidates(self, criteria, candidates, find_context, find_parameters):
        """
        Find element in given candidates with given criteria
        
        This default implementation use find_elements_by_candidates.
        It is usually recommended to override it with a dedicated implementation.
        @param candidates List of candidates
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        # Manage no candidates
        if len(candidates) == 0:
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException("No candidates")
            else:
                return None
        
        found_elements = self.find_elements_by_candidates(criteria, candidates, find_context, find_parameters)
        return self._get_element_from_list(found_elements, len(candidates), criteria, find_context, find_parameters)
    
    def find_elements(self, criteria, find_context, find_parameters):
        """
        Find elements in context container with given criteria
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found elements
        """
        if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
            return self.find_elements_by_container(criteria, find_context.container, find_context, find_parameters)
        elif isinstance(find_context, ListContainersFindContext) and find_context.nb_containers > 0:
            return self.find_elements_by_candidates(criteria, find_context.containers, find_context, find_parameters)
        else:
            raise TechnicalException("Find context has no container")
    
    def find_elements_by_container(self, criteria, container, find_context, find_parameters):
        """
        Find elements in given container with given criteria
        @param criteria Criteria
        @param container Container
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found elements
        """
        raise NotImplementedError

    def find_elements_by_candidates(self, criteria, candidates, find_context, find_parameters):
        """
        Find elements in given candidates with given criteria
        
        This default implementation use find_elements_by_container.
        It is usually recommended to override it with a dedicated implementation.
        @param candidates List of candidates
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found elements
        """
        if criteria is None:
            return candidates
        else:
            res = []
            if find_parameters.add_valid_container:
                res = self.validate_elements(criteria, candidates, find_context, find_parameters)
                
            # Continue if max number of elements is not reached
            if find_parameters.nb_max_elements is None or len(res) < find_parameters.nb_max_elements:
                sub_find_parameters = find_parameters.with_valid_container(False)
                for el in candidates:
                    found_elements = self.find_elements_by_container(criteria, el, find_context, sub_find_parameters)
                    res.extend(found_elements)
                    
                    # Stop if max number of elements is reached
                    if find_parameters.nb_max_elements is not None and len(res) >= find_parameters.nb_max_elements:
                        break
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"find_elements_by_candidates: criteria={criteria} ; nb candidates={len(candidates)} ; max nb={find_parameters.nb_max_elements} ; visibility={find_parameters.visibility}   ->  {len(res)} elements")
            return res

    def validate_elements(self, criteria, elements, find_context, find_parameters):
        """
        @param criteria Criteria
        @param elements List of elements to validate
        @param find_context Find context
        @param find_parameters Find parameters
        @return List of elements validating given criteria
        """
        if criteria is None:
            return elements
        else:
            res = []
            criteria_context = find_context.get_criteria_context()
            criteria_parameters = find_parameters.get_criteria_parameters()

            # Search elements
            for el in elements:
                if criteria.validate(el, criteria_context, criteria_parameters):
                    res.add(el)
                       
                    # Stop if max number of elements is reached
                    if criteria_parameters.nb_max_elements is not None and len(res) >= criteria_parameters.nb_max_elements:
                        break
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Validate {len(elements)} elements: criteria={criteria} ; nbMax={criteria_parameters.nb_max_elements} ; visibility={criteria_parameters.visibility}   ->  {len(res)} elements")
            return res
    
    def _get_element_from_list(self, found_elements, nb_candidates, criteria, find_context, find_parameters):
        """
        Get element from found elements that should contain only one element.
        @param found_elements List of elements previously found
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        # Analyze search result
        if len(found_elements) == 1:
            return found_elements[0]
        elif len(found_elements) > 1:
            msg_list = ["More than one"]
            if find_parameters.nb_max_elements is not None:
                msg_list.append(f" (at least {len(found_elements)})")
            else:
                msg_list.append(f" ({len(found_elements)})")
            msg_list.append(f" element was found")
            if nb_candidates > 1:
                msg_list.append(f" (in {nb_candidates} candidates)")
            if criteria is not None:
                msg_list.append(f" with criteria={criteria}")
            msg_list.append(":\n")
            msg_list.append(self.represent_elements(found_elements, 4))
            raise TooManyElementsException("".join(msg_list))
        elif find_parameters.raise_no_such_element:
            msg_list = ["Unable to find element"]
            if nb_candidates > 1:
                msg_list.append(f" (in {nb_candidates} candidates)")
            if criteria is not None:
                msg_list.append(f" with criteria={criteria}")
            raise NoSuchElementException("".join(msg_list))
    
    def represent_elements(self, elements, indent):
        """
        @param elements Elements
        @param indent Indent
        @return String representing elements
        """
        res_list = []

        for i, el in enumerate(elements):
            res_list.append(f"{i} : {self.get_element_description(el)}")
        
        return Tools.indent_string(indent, "\n".join(res_list))
    
    def get_element_description(self, element):
        """
        @param element Element
        @return Element description
        """
        raise NotImplementedError
    
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.actors.actions import ActionOnInput
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.verify_exception import VerifyException

logger = logging.getLogger(__name__)



class VerifyInput(ActionOnInput):
    """ Generic verify of input.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, raise_exception):
        super().__init__(name)
        self.__raise_exception = raise_exception
        
    @property
    def name(self):
        return self.__name
    
    def execute(self, input_):
        res = None
        try:
            res = self._verify(input_)
        except Exception as exc:
            if self.__raise_exception:
                raise exc
            else:
                res = False
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Verify [{self.name}] has failed: {Tools.represent_exception(exc)}")
                
        if not res and self.__raise_exception:
            self._raise_verify_exception()
        return res

    def _verify(self, input_):
        """
        @param input_ Input
        @return Result of verify
        """
        raise NotImplementedError

    def _raise_verify_exception(self):
        """
        Raise verify exception
        """
        raise VerifyException(f"Verify [{self.name}] has failed")
    
    
    
    
    
    